# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from inspect import signature
from typing import Dict, List, Optional

from mmengine.model import BaseModel
from torch import nn

from mmrazor.registry import MODELS
from ..algorithms.base import LossResults
from ..task_modules import DistillDeliveryManager, RecorderManager
from .base_distiller import BaseDistiller


@MODELS.register_module()
class ConfigurableDistiller(BaseDistiller):
    """``ConfigurableDistiller`` is a powerful tool that can reproduce most
    distillation algorithms without modifying the code of teacher or student
    models.

    ``ConfigurableDistiller`` can get various intermediate results of the
    model in a hacky way by ``Recorder``. More details see user-docs for
    ``Recorder``.

    ``ConfigurableDistiller`` can use the teacher's intermediate results to
    override the student's intermediate results in a hacky way by ``Delivery``.
    More details see user-docs for ``Delivery``.

    Args:
        student_recorders (dict, optional): Config for multiple recorders. A
            student model may have more than one recorder. These recorders
            only record the student model's intermediate results. Defaults to
            None.
        teacher_recorders (dict, optional): Config for multiple recorders. A
            teacher model may have more than one recorder. These recorders
            only record the teacher model's intermediate results. Defaults to
            None.
        distill_deliveries (dict, optional): Config for multiple deliveries. A
            distill algorithm may have more than one delivery. Defaults to
            None.
        connectors (dict, optional): Config for multiple connectors. A
            distillation model may have more than one connector. Defaults to
            None.
        distill_losses: (Dict[str, Dict], optional): Config for multiple
            distill losses. A distill algorithm may have more than one distill
            loss. Defaults to None.
        loss_forward_mappings: (Dict[str, Dict], optional): Mapping between
            distill loss forward arguments and records.

    Note:
        If a distill loss needs to backward, the name of the loss must contain
        "loss". If it is only used as a statistical value, the name can not
        contain "loss". More details see docs for
        :func:`mmengine.model.BaseModel._parse_loss`.

    Note:
        The keys of ``loss_forward_mappings`` should be consistent with the
        keys of ``distill_losses``.

        Each item in ``loss_forward_mappings`` is a mapping between a distill
        loss and its forward arguments. The keys of the mapping are the
        signature of the loss's forward, and the values of the mapping are the
        recorded data location.

        ``from_recorder``refers to the recorder where the data is stored, and
        if ``from_student`` is True, it means the recorder is in `
        `student_recorders``; otherwise, it means the recorder is in
        ``teacher_recorders``.

        A connector can be called according to its `connector_name`, so that a
        input can use a different connector in different loss.

    Examples:
        >>> distill_losses = dict(
        ...     loss_neck=dict(type='L2Loss', loss_weight=5))

        >>> student_recorders = dict(
        ...     feat = dict(type='ModuleOutputs', sources=['neck.gap']))

        >>> teacher_recorders = dict(
        ...     feat = dict(type='ModuleOutputs', sources=['neck.gap']))

        >>> connectors = dict(
        ...     loss_neck_sfeat = dict(
        ...         type='SingleConvConnector', in_channel=32, out_channel=64),
        ...     loss_neck_tfeat = dict(
        ...         type='SingleConvConnector', in_channel=32, out_channel=64))

        >>> loss_forward_mappings = dict(
        ...     loss_neck=dict(
        ...         s_feature=dict(from_recorder='feat', from_student=True,
        ...                        connector='loss_neck_sfeat'),
        ...         t_feature=dict(from_recorder='feat', from_student=False,
        ...                        connector='loss_neck_tfeat')))
    """

    def __init__(self,
                 student_recorders: Optional[Dict[str, Dict]] = None,
                 teacher_recorders: Optional[Dict[str, Dict]] = None,
                 distill_deliveries: Optional[Dict[str, Dict]] = None,
                 connectors: Optional[Dict[str, Dict]] = None,
                 distill_losses: Optional[Dict[str, Dict]] = None,
                 loss_forward_mappings: Optional[Dict[str, Dict]] = None,
                 **kwargs):
        super().__init__(**kwargs)
        # The recorder manager is just constructed, but not really initialized
        # yet. Recorder manager initialization needs to input the corresponding
        # model.
        self.student_recorders = RecorderManager(student_recorders)
        self.teacher_recorders = RecorderManager(teacher_recorders)

        self.deliveries = DistillDeliveryManager(distill_deliveries)

        self.distill_losses = self.build_distill_losses(distill_losses)

        self.connectors = self.build_connectors(connectors)

        if loss_forward_mappings:
            # Check if loss_forward_mappings is in the correct format.
            self._check_loss_forward_mappings(self.distill_losses,
                                              loss_forward_mappings,
                                              self.student_recorders,
                                              self.teacher_recorders)
            self.loss_forward_mappings = loss_forward_mappings
        else:
            self.loss_forward_mappings = dict()

    def set_deliveries_override(self, override: bool) -> None:
        """Set the `override_data` of all deliveries."""
        self.deliveries.override_data = override

    def prepare_from_student(self, model: BaseModel) -> None:
        """Initialize student recorders."""
        self.student_recorders.initialize(model)

    def prepare_from_teacher(self, model: nn.Module) -> None:
        """Initialize teacher recorders."""
        self.teacher_recorders.initialize(model)

    def build_connectors(
        self,
        connectors: Optional[Dict[str, Dict]] = None,
    ) -> nn.ModuleDict:
        """Initialize connectors."""

        distill_connecotrs = nn.ModuleDict()
        if connectors:
            for connector_name, connector_cfg in connectors.items():
                connector = MODELS.build(connector_cfg)
                distill_connecotrs[connector_name] = connector

        return distill_connecotrs

    def build_distill_losses(
        self,
        losses: Optional[Dict[str, Dict]] = None,
    ) -> nn.ModuleDict:
        """build distill losses according config."""

        distill_losses = nn.ModuleDict()
        if losses:
            for loss_name, loss_cfg in losses.items():
                assert loss_name not in distill_losses
                if 'loss' not in loss_name:
                    warnings.warn(
                        f'Warning: If {loss_name} is a loss that needs to '
                        f'backward, the name of {loss_name} must contain '
                        f'"loss". If it is only used as a statistical value, '
                        'then the name must not contain "loss". More details '
                        'see docs for '
                        ':func:`mmengine.model.BaseModel._parse_loss`',
                        UserWarning)
                item_loss = MODELS.build(loss_cfg)
                distill_losses[loss_name] = item_loss

        return distill_losses

    def get_record(self,
                   recorder: str,
                   from_student: bool,
                   record_idx: int = 0,
                   data_idx: Optional[int] = None,
                   connector: Optional[str] = None,
                   connector_idx: Optional[int] = None) -> List:
        """According to each item in ``record_infos``, get the corresponding
        record in ``recorder_manager``."""

        if from_student:
            recorder_ = self.student_recorders.get_recorder(recorder)
        else:
            recorder_ = self.teacher_recorders.get_recorder(recorder)
        record_data = recorder_.get_record_data(record_idx, data_idx)

        if connector:
            record_data = self.connectors[connector](record_data)
        if connector_idx is not None:
            record_data = record_data[connector_idx]

        return record_data

    def compute_distill_losses(self) -> LossResults:
        """Compute distill losses automatically."""
        # Record all computed losses' results.
        losses = dict()
        for loss_name, forward_mappings in self.loss_forward_mappings.items():
            forward_kwargs = dict()
            for forward_key, record in forward_mappings.items():
                forward_var = self.get_record(**record)
                forward_kwargs[forward_key] = forward_var

            loss_module = self.distill_losses[loss_name]
            loss = loss_module(**forward_kwargs)  # type: ignore
            # add computed loss result.
            losses[loss_name] = loss

        return losses

    def _check_loss_forward_mappings(
            self, losses: nn.ModuleDict, loss_forward_mappings: Dict[str,
                                                                     Dict],
            student_recorders: RecorderManager,
            teacher_recorders: RecorderManager) -> None:
        """Check if ``loss_forward_mappings`` is in the correct format."""

        if not isinstance(loss_forward_mappings, dict):
            raise TypeError(
                'loss_forward_mappings should be a dict instance, but got'
                f'{type(loss_forward_mappings)}')

        for loss_name, forward_mappings in loss_forward_mappings.items():
            assert loss_name in losses, \
                f'"{loss_name}" is not in distill losses. The keys of ' \
                'loss_forward_kwargs must match the keys of distill_losses.'

            if not isinstance(forward_mappings, dict):
                raise TypeError(
                    'Each item of loss_forward_mappings should be a dict '
                    f'instance, but got {type(forward_mappings)}')

            loss_module = losses[loss_name]
            loss_forward_params = signature(loss_module.forward).parameters
            loss_forward_keys = loss_forward_params.keys()
            # Allow default params.
            # Check non-default params, not len(params).

            for forward_key, record_info in forward_mappings.items():
                assert forward_key in loss_forward_keys, \
                    f'{forward_key} is not in the signature of \
                    {type(loss_module).__name__} forward, \
                    please check your config.'

                if (loss_forward_params[forward_key].default !=
                        loss_forward_params[forward_key].empty):
                    # default params without check
                    continue

                assert 'recorder' in record_info, \
                    'Each item of loss_forward_mappings should have ' \
                    '"recorder", pls check your config.'

                assert 'from_student' in record_info, \
                    'Each item of loss_forward_mappings should have ' \
                    '"from_student", pls check your config.'

                recorder: str = record_info['recorder']
                from_student: bool = record_info['from_student']

                if not isinstance(from_student, bool):
                    raise TypeError(f'from_student should be a bool instance, '
                                    f'but got {type(from_student)}')

                if from_student:
                    assert recorder in student_recorders.recorders, \
                        f'For {forward_key}, "{recorder}" must be in \
                        `student_recorders`.'

                else:
                    assert recorder in teacher_recorders.recorders, \
                        f'For {forward_key}, "{recorder}" must be in \
                        `teacher_recorders`.'

                if 'connector' in record_info:
                    connector: str = record_info['connector']
                    assert connector in self.connectors, \
                        f'{connector} must be in "connectors".'



_base_ = 'imagenet_bs2048_autoslim.py'

_RandomResizedCrop_cfg = _base_.train_dataloader.dataset.pipeline[1]
assert _RandomResizedCrop_cfg.type == 'mmcls.RandomResizedCrop'
_RandomResizedCrop_cfg.backend = 'pillow'

_ResizeEdge_cfg = _base_.test_dataloader.dataset.pipeline[1]
assert _ResizeEdge_cfg.type == 'mmcls.ResizeEdge'
_ResizeEdge_cfg.backend = 'pillow'



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Sequence, Union

import torch
from mmengine.evaluator import Evaluator
from mmengine.runner import ValLoop, autocast
from torch.utils.data import DataLoader

from mmrazor.registry import LOOPS


@LOOPS.register_module()
class SingleTeacherDistillValLoop(ValLoop):
    """Knowledge Distill loop for validation. It is not only validate student,
    but also validate teacher with the same dataloader.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict): A dataloader object or a dict to
            build a dataloader.
        evaluator (Evaluator or dict or list): Used for computing metrics.
        fp16 (bool): Whether to enable fp16 validation. Defaults to
            False.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[DataLoader, Dict],
                 evaluator: Union[Evaluator, Dict, List],
                 fp16: bool = False) -> None:
        super().__init__(runner, dataloader, evaluator, fp16)
        if self.runner.distributed:
            assert hasattr(self.runner.model.module, 'teacher')
            # TODO: remove hard code after mmcls add data_preprocessor
            data_preprocessor = self.runner.model.module.data_preprocessor
            self.teacher = self.runner.model.module.teacher
            self.teacher.data_preprocessor = data_preprocessor

        else:
            assert hasattr(self.runner.model, 'teacher')
            # TODO: remove hard code after mmcls add data_preprocessor
            data_preprocessor = self.runner.model.data_preprocessor
            self.teacher = self.runner.model.teacher
            self.teacher.data_preprocessor = data_preprocessor

    def run(self):
        """Launch validation."""
        self.runner.call_hook('before_val')
        self.runner.call_hook('before_val_epoch')
        self.runner.model.eval()

        for idx, data_batch in enumerate(self.dataloader):
            self.run_iter(idx, data_batch)
        # compute student metrics
        metrics = self.evaluator.evaluate(len(self.dataloader.dataset))
        student_metrics = dict()
        for key, value in metrics.items():
            student_key = 'student.' + key
            teacher_key = 'teacher.' + key

            student_metrics[student_key] = value
            self.runner.message_hub.log_scalars.pop(f'val/{teacher_key}', None)

        self.runner.call_hook('after_val_epoch', metrics=student_metrics)

        self.runner.call_hook('before_val_epoch')
        for idx, data_batch in enumerate(self.dataloader):
            self.run_iter_teacher(idx, data_batch)
        # compute teacher metrics
        metrics = self.evaluator.evaluate(len(self.dataloader.dataset))
        teacher_metrics = dict()
        for key, value in metrics.items():
            student_key = 'student.' + key
            teacher_key = 'teacher.' + key

            teacher_metrics[teacher_key] = value
            self.runner.message_hub.log_scalars.pop(f'val/{student_key}', None)

        self.runner.call_hook('after_val_epoch', metrics=teacher_metrics)
        self.runner.call_hook('after_val')

    @torch.no_grad()
    def run_iter_teacher(self, idx, data_batch: Sequence[dict]):
        """Iterate one mini-batch.

        Args:
            data_batch (Sequence[dict]): Batch of data
                from dataloader.
        """
        self.runner.call_hook(
            'before_val_iter', batch_idx=idx, data_batch=data_batch)

        with autocast(enabled=self.fp16):
            # outputs should be sequence of BaseDataElement
            outputs = self.teacher.val_step(data_batch)

        self.evaluator.process(data_samples=outputs, data_batch=data_batch)
        self.runner.call_hook(
            'after_val_iter',
            batch_idx=idx,
            data_batch=data_batch,
            outputs=outputs)


@LOOPS.register_module()
class SelfDistillValLoop(ValLoop):
    """Knowledge Distill loop for validation. Only validate student.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict): A dataloader object or a dict to
            build a dataloader.
        evaluator (Evaluator or dict or list): Used for computing metrics.
        fp16 (bool): Whether to enable fp16 validation. Defaults to
            False.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[DataLoader, Dict],
                 evaluator: Union[Evaluator, Dict, List],
                 fp16: bool = False) -> None:
        super().__init__(runner, dataloader, evaluator, fp16)

    def run(self):
        """Launch validation."""
        self.runner.call_hook('before_val')
        self.runner.call_hook('before_val_epoch')
        self.runner.model.eval()

        for idx, data_batch in enumerate(self.dataloader):
            self.run_iter(idx, data_batch)
        # compute student metrics
        metrics = self.evaluator.evaluate(len(self.dataloader.dataset))
        student_metrics = dict()
        for key, value in metrics.items():
            student_key = 'student.' + key
            student_metrics[student_key] = value

        self.runner.call_hook('after_val_epoch', metrics=student_metrics)
        self.runner.call_hook('after_val')



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS

try:
    from mmcls.models.losses.cross_entropy_loss import soft_cross_entropy
except ImportError:
    from mmrazor.utils import get_placeholder
    soft_cross_entropy = get_placeholder('mmcls')


@MODELS.register_module()
class KDSoftCELoss(nn.Module):
    """Distilling the Knowledge in a Neural Network, NIPS2014. Based on Soft
    Cross Entropy criterion.

    https://arxiv.org/pdf/1503.02531.pdf


    Args:
        tau (int, optional): Temperature. Defaults to 1.
        reduction (str): Specifies the reduction to apply to the loss:
            ``'none'`` | ``'none'`` | ``'sum'`` | ``'mean'``.
            ``'none'``: no reduction will be applied,
            ``'sum'``: the output will be summed,
            ``'mean'``: the output will be divided by the number of
                elements in the output.
            Default: ``'mean'``
        mult_tem_square (bool, optional): Multiply square of temperature
            or not. Defaults to True.
        loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(
        self,
        tau: float = 1.0,
        reduction: str = 'mean',
        mult_tem_square: bool = True,
        loss_weight: float = 1.0,
    ) -> None:
        super().__init__()
        self.tau = tau
        self.mult_tem_square = mult_tem_square
        self.loss_weight = loss_weight
        self.cls_criterion = soft_cross_entropy

        accept_reduction = {None, 'none', 'mean', 'sum'}
        assert reduction in accept_reduction, \
            f'KLDivergence supports reduction {accept_reduction}, ' \
            f'but gets {reduction}.'
        self.reduction = reduction

    def forward(
        self,
        preds_S: torch.Tensor,
        preds_T: torch.Tensor,
        weight: torch.Tensor = None,
        avg_factor: int = None,
        reduction_override: str = None,
    ) -> torch.Tensor:
        """Forward computation.

        Args:
            preds_S (torch.Tensor): The student model prediction with
                shape (N, C).
            preds_T (torch.Tensor): The teacher model prediction with
                shape (N, C).
            weight (torch.Tensor, optional): Sample-wise loss weight with
                shape (N, C). Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optiom): Override redunction in forward.
                Defaults to None.

        Return:
            torch.Tensor: The calculated loss value.
        """
        reduction = (
            reduction_override if reduction_override else self.reduction)

        preds_S = preds_S / self.tau
        soft_label = F.softmax((preds_T / self.tau), dim=-1)
        loss_cls = self.loss_weight * self.cls_criterion(
            preds_S,
            soft_label,
            weight,
            reduction=reduction,
            avg_factor=avg_factor)
        if self.mult_tem_square:
            loss_cls *= (self.tau**2)
        return loss_cls



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractclassmethod


class BaseCounter(object, metaclass=ABCMeta):
    """Base class of all op module counters in `TASK_UTILS`.

    In ResourceEstimator, `XXModuleCounter` is responsible for `XXModule`,
    which refers to estimator/flops_params_counter.py::get_counter_type().
    Users can customize a `ModuleACounter` and overwrite the `add_count_hook`
    method with a self-defined module `ModuleA`.
    """

    def __init__(self) -> None:
        pass

    @staticmethod
    @abstractclassmethod
    def add_count_hook(module, input, output):
        """The main method of a `BaseCounter` which defines the way to
        calculate resources(flops/params) of the current module.

        Args:
            module (nn.Module): the module to be tested.
            input (_type_): input_tensor. Plz refer to `torch forward_hook`
            output (_type_): output_tensor. Plz refer to `torch forward_hook`
        """
        pass



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS


@MODELS.register_module()
class ChannelWiseDivergence(nn.Module):
    """PyTorch version of `Channel-wise Distillation for Semantic Segmentation.

    <https://arxiv.org/abs/2011.13256>`_.

    Args:
        tau (float): Temperature coefficient. Defaults to 1.0.
        loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(
        self,
        tau=1.0,
        loss_weight=1.0,
    ):
        super(ChannelWiseDivergence, self).__init__()
        self.tau = tau
        self.loss_weight = loss_weight

    def forward(self, preds_S, preds_T):
        """Forward computation.

        Args:
            preds_S (torch.Tensor): The student model prediction with
                shape (N, C, H, W).
            preds_T (torch.Tensor): The teacher model prediction with
                shape (N, C, H, W).

        Return:
            torch.Tensor: The calculated loss value.
        """
        assert preds_S.shape[-2:] == preds_T.shape[-2:]
        N, C, H, W = preds_S.shape

        softmax_pred_T = F.softmax(preds_T.view(-1, W * H) / self.tau, dim=1)

        logsoftmax = torch.nn.LogSoftmax(dim=1)
        loss = torch.sum(softmax_pred_T *
                         logsoftmax(preds_T.view(-1, W * H) / self.tau) -
                         softmax_pred_T *
                         logsoftmax(preds_S.view(-1, W * H) / self.tau)) * (
                             self.tau**2)

        loss = self.loss_weight * loss / (C * N)

        return loss



# Copyright (c) OpenMMLab. All rights reserved.
from .delivery import *  # noqa: F401,F403
from .estimators import ResourceEstimator
from .recorder import *  # noqa: F401,F403
from .tracer import *  # noqa: F401,F403

__all__ = ['ResourceEstimator']



# Copyright (c) OpenMMLab. All rights reserved.
import math
from operator import attrgetter

import torch
import torch.nn as nn
from scipy.stats import norm

from mmrazor.registry import MODELS
from ..architectures.connectors import OFDTeacherConnector
from ..losses import OFDLoss
from .configurable_distiller import ConfigurableDistiller


@MODELS.register_module()
class OFDDistiller(ConfigurableDistiller):
    """Distiller for ``OverhaulFeatureDistillation``, inherited from
    ``ConfigurableDistiller``, add func:

    ``init_ofd_connectors`` to initialize margin.
    """

    def init_ofd_connectors(self, teacher: nn.Module) -> None:
        """Initialize OFD connectors' `margin`."""
        for loss_key, loss_forward_mapping in self.loss_forward_mappings.items(
        ):
            if isinstance(self.distill_losses[loss_key], OFDLoss):
                for _input_keys, _input_mapping in loss_forward_mapping.items(
                ):
                    recorder_mgn = self.student_recorders if _input_mapping[
                        'from_student'] else self.teacher_recorders
                    recorder = recorder_mgn.get_recorder(
                        _input_mapping['recorder'])
                    module_key = recorder.source
                    bn_module = attrgetter(module_key)(teacher)

                    assert isinstance(
                        bn_module, (nn.BatchNorm2d, nn.SyncBatchNorm)), (
                            'Overhaul distillation only support connection on '
                            'layers: [`BatchNorm2d`, `SyncBatchNorm`]')

                    if 'connector' in _input_mapping and not _input_mapping[
                            'from_student']:
                        connector = self.connectors[
                            _input_mapping['connector']]
                        assert isinstance(connector, OFDTeacherConnector), (
                            'OFD loss mapping for `t_feature` expect type '
                            '`OFDTeacherConnector`, but get '
                            f'`{type(connector)}`')
                        margin = self._get_margin_from_BN(bn_module)
                        connector.init_margin(margin)

    def _get_margin_from_BN(self, bn: nn.BatchNorm2d) -> torch.Tensor:
        """Get margin from BN layer.

        Args:
            bn (nn.BatchNorm2d): input module, must be a BN layer.

        Returns:
            torch.Tensor: margin
        """
        margin = []
        std = bn.weight.data
        mean = bn.bias.data
        for (s, m) in zip(std, mean):
            s = abs(s.item())
            m = m.item()
            if norm.cdf(-m / s) > 0.001:
                margin.append(-s * math.exp(-(m / s)**2 / 2) /
                              math.sqrt(2 * math.pi) / norm.cdf(-m / s) + m)
            else:
                margin.append(-3 * s)
        return torch.FloatTensor(margin).unsqueeze(1).unsqueeze(2).unsqueeze(
            0).detach()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Optional, Tuple

from torch import nn

from mmrazor.registry import TASK_UTILS
from .base_recorder import BaseRecorder


@TASK_UTILS.register_module()
class ModuleOutputsRecorder(BaseRecorder):
    """Recorder for intermediate results which are Pytorch moudle's outputs.

    Examples:
            >>> from torch import nn
            >>> class ToyModel(nn.Module):
            ...     def __init__(self):
            ...         super().__init__()
            ...         self.conv1 = nn.Conv2d(8,8,1)
            ...         self.conv2 = nn.Conv2d(1,1,1)
            ...     def forward(self, x):
            ...         x1 = self.conv1(x)
            ...         x2 = self.conv1(x+1)
            ...         return self.conv2(x1 + x2)

            >>> model = ToyModel()
            >>> [ name for name,_ in model.named_modules() ]
            ['conv1', 'conv2']

            >>> r1 = ModuleOutputsRecorder('conv1')
            >>> r1.initialize(model)

            >>> with r1:
            >>>     res = model(torch.randn(1,1,1,1))

            >>> r1.data_buffer
            [tensor([[[[0.6734]]]]), tensor([[[[1.2514]]]]) ]
            >>> r1.get_record_data(record_idx=1)
            tensor([[[[1.2514]]]])

            >>> r2 = ModuleOutputsRecorder('conv2')
            >>> r2.initialize(model)

            >>> with r2:
            >>>     res = model(torch.randn(1,1,1,1))

            >>> r2.data_buffer
            [tensor([[[[0.9534]]]])]
            >>> r2.get_record_data()
            tensor([[[[0.9534]]]])
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._recording = False

    @property
    def recording(self) -> bool:
        """bool: whether to record data in forward hook."""
        return self._recording

    def prepare_from_model(self, model: Optional[nn.Module] = None) -> None:
        """Register Pytorch forward hook to corresponding module."""

        assert model is not None, 'model can not be None.'

        founded = False
        for name, module in model.named_modules():
            if name == self.source:
                module.register_forward_hook(self.forward_hook)
                founded = True
                break

        assert founded, f'"{self.source}" is not in the model.'

    def forward_hook(self, module: nn.Module, inputs: Tuple,
                     outputs: Any) -> None:
        """Save the module's forward output.

        Args:
            module (:obj:`torch.nn.Module`): The module to register hook.
            inputs (tuple): The input of the module.
            outputs : The output of the module.
        """
        if self._recording:
            self.data_buffer.append(outputs)

    def __enter__(self):
        """Enter the context manager."""
        super().__enter__()
        self._recording = True

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        super().__exit__(exc_type, exc_value, traceback)
        self._recording = False



# Copyright (c) OpenMMLab. All rights reserved.
from math import log
from typing import Dict, Optional, Tuple, Union

import torch
import torch.nn as nn

from mmrazor.registry import MODELS
from ..ops.darts_series import DartsSepConv
from .base_connector import BaseConnector


@MODELS.register_module()
class BYOTConnector(BaseConnector):
    """BYOTConnector connector that adds a self-attention with DartsSepConv.

    Args:
        in_channel (int): The input channel of the DartsSepConv.
            Use like input_tensor_channel = in_channel * expansion.
        out_channel (int): The output channel of the DartsSepConv.
            Use like output_tensor_channel = out_channel * expansion.
        num_classes (int): The classification class num.
        expansion (int): Expansion of DartsSepConv. Default to 4.
        pool_size (int | tuple[int]): Average 2D pool size. Default to 4.
        kernel_size (int | tuple[int]): Size of the convolving kernel in
            DartsSepConv. Same as that in ``nn._ConvNd``. Default to 3.
        stride (int | tuple[int]): Stride of the first layer in DartsSepConv.
            Same as that in ``nn._ConvNd``. Default to 1.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(
        self,
        in_channel: int,
        out_channel: int,
        num_classes: int,
        expansion: int = 4,
        pool_size: Union[int, Tuple[int]] = 4,
        kernel_size: Union[int, Tuple[int]] = 3,
        stride: Union[int, Tuple[int]] = 1,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(init_cfg)
        self.attention = nn.Sequential(
            DartsSepConv(
                in_channels=in_channel * expansion,
                out_channels=in_channel * expansion,
                kernel_size=kernel_size,
                stride=stride), nn.BatchNorm2d(in_channel * expansion),
            nn.ReLU(), nn.Upsample(scale_factor=2, mode='bilinear'),
            nn.Sigmoid())
        scala_num = log(out_channel / in_channel, 2)
        assert scala_num.is_integer()
        scala = []

        _in_channel = in_channel

        for _ in range(int(scala_num)):
            scala.append(
                DartsSepConv(
                    in_channels=_in_channel * expansion,
                    out_channels=_in_channel * 2 * expansion,
                    kernel_size=kernel_size,
                    stride=stride))
            _in_channel *= 2
        scala.append(nn.AvgPool2d(pool_size))
        self.scala = nn.Sequential(*scala)
        self.fc = nn.Linear(out_channel * expansion, num_classes)

    def forward_train(self, feature: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Forward computation.

        Args:
            feature (torch.Tensor): Input feature.
        """
        feat = self.attention(feature)
        feat = feat * feature

        feat = self.scala(feat)
        feat = feat.view(feature.size(0), -1)
        logits = self.fc(feat)
        return (feat, logits)



# Copyright (c) OpenMMLab. All rights reserved.
# yapf: disable
from .default_module_converters import (DEFAULT_MODULE_CONVERTERS,
                                        dynamic_bn_converter,
                                        dynamic_conv2d_converter,
                                        dynamic_gn_converter,
                                        dynamic_in_converter,
                                        dynamic_linear_converter)
# yapf: enable
from .slimmable_bn_converter import switchable_bn_converter

__all__ = [
    'dynamic_conv2d_converter', 'dynamic_linear_converter',
    'dynamic_bn_converter', 'dynamic_in_converter', 'dynamic_gn_converter',
    'DEFAULT_MODULE_CONVERTERS', 'switchable_bn_converter'
]



_base_ = [
    'mmrazor::_base_/settings/imagenet_bs2048_autoslim_pil.py',
    'mmcls::_base_/models/mobilenet_v2_1x.py',
    'mmcls::_base_/default_runtime.py',
]

supernet = _base_.model
supernet.backbone.widen_factor = 1.5
supernet.head.in_channels = 1920

# !dataset config
# ==========================================================================
# data preprocessor
data_preprocessor = dict(
    type='ImgDataPreprocessor',
    # RGB format normalization parameters
    mean=[123.675, 116.28, 103.53],
    std=[58.395, 57.12, 57.375],
    # convert image from BGR to RGB
    bgr_to_rgb=True,
)

# !autoslim algorithm config
num_samples = 2
model = dict(
    _delete_=True,
    _scope_='mmrazor',
    type='AutoSlim',
    num_samples=num_samples,
    architecture=supernet,
    data_preprocessor=data_preprocessor,
    distiller=dict(
        type='ConfigurableDistiller',
        teacher_recorders=dict(
            fc=dict(type='ModuleOutputs', source='head.fc')),
        student_recorders=dict(
            fc=dict(type='ModuleOutputs', source='head.fc')),
        distill_losses=dict(
            loss_kl=dict(type='KLDivergence', tau=1, loss_weight=1)),
        loss_forward_mappings=dict(
            loss_kl=dict(
                preds_S=dict(recorder='fc', from_student=True),
                preds_T=dict(recorder='fc', from_student=False)))),
    mutator=dict(
        type='OneShotChannelMutator',
        mutable_cfg=dict(
            type='OneShotMutableChannel',
            candidate_choices=list(i / 12 for i in range(2, 13)),
            candidate_mode='ratio'),
        tracer_cfg=dict(
            type='BackwardTracer',
            loss_calculator=dict(type='ImageClassifierPseudoLoss'))))

model_wrapper_cfg = dict(
    type='mmrazor.AutoSlimDDP',
    broadcast_buffers=False,
    find_unused_parameters=False)

optim_wrapper = dict(accumulative_counts=num_samples + 2)

# learning policy
max_epochs = 50
param_scheduler = dict(end=max_epochs)

# train, val, test setting
train_cfg = dict(max_epochs=max_epochs)
val_cfg = dict(type='mmrazor.AutoSlimValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
import os
from typing import Dict, List, Optional, Union

import torch
from mmengine.model import BaseModel, MMDistributedDataParallel
from mmengine.optim import OptimWrapper, OptimWrapperDict
from torch import nn
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.mutators import DiffModuleMutator
from mmrazor.models.utils import add_prefix
from mmrazor.registry import MODEL_WRAPPERS, MODELS
from mmrazor.utils import FixMutable
from ..base import BaseAlgorithm


@MODELS.register_module()
class Darts(BaseAlgorithm):
    """Implementation of `DARTS <https://arxiv.org/abs/1806.09055>`_

    DARTS means Differentiable Architecture Search, a classic NAS algorithm.
    :class:`Darts` implements the APIs required by the DARTS, as well as the
    supernet training and subnet retraining logic for each iter.

    Args:
        architecture (dict|:obj:`BaseModel`): The config of :class:`BaseModel`
            or built model. Corresponding to supernet in NAS algorithm.
        mutator (dict|:obj:`DiffModuleMutator`): The config of
            :class:`DiffModuleMutator` or built mutator.
        fix_subnet (str | dict | :obj:`FixSubnet`): The path of yaml file or
            loaded dict or built :obj:`FixSubnet`.
        norm_training (bool): Whether to set norm layers to training mode,
            namely, not freeze running stats (mean and var). Note: Effect on
            Batch Norm and its variants only. Defaults to False.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Defaults to None.
        init_cfg (dict): Init config for ``BaseModule``.

    Note:
        Darts has two training mode: supernet training and subnet retraining.
        If `fix_subnet` is None, it means supernet training.
        If `fix_subnet` is not None, it means subnet training.
    """

    def __init__(self,
                 architecture: Union[BaseModel, Dict],
                 mutator: Optional[Union[DiffModuleMutator, Dict]] = None,
                 fix_subnet: Optional[FixMutable] = None,
                 unroll: bool = False,
                 norm_training: bool = False,
                 data_preprocessor: Optional[Union[dict, nn.Module]] = None,
                 init_cfg: Optional[dict] = None):
        super().__init__(architecture, data_preprocessor, init_cfg)

        # Darts has two training mode: supernet training and subnet retraining.
        # fix_subnet is not None, means subnet retraining.
        if fix_subnet:
            # Avoid circular import
            from mmrazor.structures import load_fix_subnet

            # According to fix_subnet, delete the unchosen part of supernet
            load_fix_subnet(self.architecture, fix_subnet)
            self.is_supernet = False
        else:
            assert mutator is not None, \
                'mutator cannot be None when fix_subnet is None.'
            if isinstance(mutator, DiffModuleMutator):
                self.mutator = mutator
            elif isinstance(mutator, dict):
                self.mutator = MODELS.build(mutator)
            else:
                raise TypeError('mutator should be a `dict` or '
                                f'`DiffModuleMutator` instance, but got '
                                f'{type(mutator)}')

            # Mutator is an essential component of the NAS algorithm. It
            # provides some APIs commonly used by NAS.
            # Before using it, you must do some preparation according to
            # the supernet.
            self.mutator.prepare_from_supernet(self.architecture)
            self.is_supernet = True

        self.norm_training = norm_training
        # TODO support unroll
        self.unroll = unroll

    def search_subnet(self):
        """Search subnet by mutator."""

        # Avoid circular import
        from mmrazor.structures import export_fix_subnet

        subnet = self.mutator.sample_choices()
        self.mutator.set_choices(subnet)
        return export_fix_subnet(self)

    def train(self, mode=True):
        """Convert the model into eval mode while keep normalization layer
        unfreezed."""

        super().train(mode)
        if self.norm_training and not mode:
            for module in self.architecture.modules():
                if isinstance(module, _BatchNorm):
                    module.training = True

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating are also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.
        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``,
                ``num_samples``.
                ``loss`` is a tensor for back propagation, which can be a
                weighted sum of multiple losses.
                ``log_vars`` contains all the variables to be sent to the
                logger.
                ``num_samples`` indicates the batch size (when the model is
                DDP, it means the batch size on each GPU), which is used for
                averaging the logs.
        """
        if isinstance(data, (tuple, list)) and isinstance(
                optim_wrapper, OptimWrapperDict):
            assert len(data) == len(optim_wrapper), \
                f'The length of data ({len(data)}) should be equal to that '\
                f'of optimizers ({len(optim_wrapper)}).'
            # TODO check the order of data
            supernet_data, mutator_data = data

            log_vars = dict()
            # TODO support unroll
            with optim_wrapper['mutator'].optim_context(self):
                mutator_batch_inputs, mutator_data_samples = \
                    self.data_preprocessor(mutator_data, True)
                mutator_loss = self(
                    mutator_batch_inputs, mutator_data_samples, mode='loss')
            mutator_losses, mutator_log_vars = self.parse_losses(mutator_loss)
            optim_wrapper['mutator'].update_params(mutator_losses)
            log_vars.update(add_prefix(mutator_log_vars, 'mutator'))

            with optim_wrapper['architecture'].optim_context(self):
                supernet_batch_inputs, supernet_data_samples = \
                    self.data_preprocessor(supernet_data, True)
                supernet_loss = self(
                    supernet_batch_inputs, supernet_data_samples, mode='loss')
            supernet_losses, supernet_log_vars = self.parse_losses(
                supernet_loss)
            optim_wrapper['architecture'].update_params(supernet_losses)
            log_vars.update(add_prefix(supernet_log_vars, 'supernet'))

        else:
            # Enable automatic mixed precision training context.
            with optim_wrapper.optim_context(self):
                batch_inputs, data_samples = self.data_preprocessor(data, True)
                losses = self(batch_inputs, data_samples, mode='loss')
            parsed_losses, log_vars = self.parse_losses(losses)
            optim_wrapper.update_params(parsed_losses)

        return log_vars


@MODEL_WRAPPERS.register_module()
class DartsDDP(MMDistributedDataParallel):

    def __init__(self,
                 *,
                 device_ids: Optional[Union[List, int, torch.device]] = None,
                 **kwargs) -> None:
        if device_ids is None:
            if os.environ.get('LOCAL_RANK') is not None:
                device_ids = [int(os.environ['LOCAL_RANK'])]
        super().__init__(device_ids=device_ids, **kwargs)

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating are also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.
        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``,
                ``num_samples``.
                ``loss`` is a tensor for back propagation, which can be a
                weighted sum of multiple losses.
                ``log_vars`` contains all the variables to be sent to the
                logger.
                ``num_samples`` indicates the batch size (when the model is
                DDP, it means the batch size on each GPU), which is used for
                averaging the logs.
        """
        if isinstance(data, (tuple, list)) and isinstance(
                optim_wrapper, OptimWrapperDict):
            assert len(data) == len(optim_wrapper), \
                f'The length of data ({len(data)}) should be equal to that '\
                f'of optimizers ({len(optim_wrapper)}).'
            # TODO check the order of data
            supernet_data, mutator_data = data

            log_vars = dict()
            # TODO process the input

            with optim_wrapper['mutator'].optim_context(self):
                mutator_batch_inputs, mutator_data_samples = \
                    self.module.data_preprocessor(mutator_data, True)
                mutator_loss = self(
                    mutator_batch_inputs, mutator_data_samples, mode='loss')
            mutator_losses, mutator_log_vars = self.module.parse_losses(
                mutator_loss)
            optim_wrapper['mutator'].update_params(mutator_losses)
            log_vars.update(add_prefix(mutator_log_vars, 'mutator'))

            with optim_wrapper['architecture'].optim_context(self):
                supernet_batch_inputs, supernet_data_samples = \
                    self.module.data_preprocessor(supernet_data, True)
                supernet_loss = self(
                    supernet_batch_inputs, supernet_data_samples, mode='loss')
            supernet_losses, supernet_log_vars = self.module.parse_losses(
                supernet_loss)
            optim_wrapper['architecture'].update_params(supernet_losses)
            log_vars.update(add_prefix(supernet_log_vars, 'supernet'))

        else:
            # Enable automatic mixed precision training context.
            with optim_wrapper.optim_context(self):
                batch_inputs, data_samples = self.module.data_preprocessor(
                    data, True)
                losses = self(batch_inputs, data_samples, mode='loss')
            parsed_losses, log_vars = self.module.parse_losses(losses)
            optim_wrapper.update_params(parsed_losses)

        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Union

from mmengine.evaluator import Evaluator
from mmengine.runner import ValLoop
from torch.utils.data import DataLoader

from mmrazor.models.utils import add_prefix
from mmrazor.registry import LOOPS


@LOOPS.register_module()
class AutoSlimValLoop(ValLoop):

    def __init__(self,
                 runner,
                 dataloader: Union[DataLoader, Dict],
                 evaluator: Union[Evaluator, Dict, List],
                 fp16: bool = False) -> None:
        super().__init__(runner, dataloader, evaluator, fp16)

        if self.runner.distributed:
            model = self.runner.model.module
        else:
            model = self.runner.model

        # just for convenience
        self._model = model

    def run(self):
        """Launch validation."""
        self.runner.call_hook('before_val')

        all_metrics = dict()

        self._model.set_max_subnet()
        metrics = self._evaluate_once()
        all_metrics.update(add_prefix(metrics, 'max_subnet'))

        self._model.set_min_subnet()
        metrics = self._evaluate_once()
        all_metrics.update(add_prefix(metrics, 'min_subnet'))

        for subnet_idx in range(self._model.num_samples):
            self._model.set_subnet(self._model.sample_subnet())
            # compute student metrics
            metrics = self._evaluate_once()
            all_metrics.update(
                add_prefix(metrics, f'random_subnet_{subnet_idx}'))

        self.runner.call_hook('after_val_epoch', metrics=all_metrics)

        self.runner.call_hook('after_val')

    def _evaluate_once(self) -> Dict:
        self.runner.call_hook('before_val_epoch')
        self.runner.model.eval()
        for idx, data_batch in enumerate(self.dataloader):
            self.run_iter(idx, data_batch)

        return self.evaluator.evaluate(len(self.dataloader.dataset))



_base_ = [
    'mmcls::_base_/datasets/cifar10_bs16.py',
    '../../../_base_/vanilla_models/wrn16_2_cifar10.py',
    'mmcls::_base_/schedules/cifar10_bs128.py',
    'mmcls::_base_/default_runtime.py',
]
test_evaluator = dict(topk=(1, 5))



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS


@MODELS.register_module()
class DKDLoss(nn.Module):
    """Decoupled Knowledge Distillation, CVPR2022.

    link: https://arxiv.org/abs/2203.08679
    reformulate the classical KD loss into two parts:
        1. target class knowledge distillation (TCKD)
        2. non-target class knowledge distillation (NCKD).
    Args:
    tau (float): Temperature coefficient. Defaults to 1.0.
    alpha (float): Weight of TCKD loss. Defaults to 1.0.
    beta (float): Weight of NCKD loss. Defaults to 1.0.
    reduction (str): Specifies the reduction to apply to the loss:
        ``'none'`` | ``'batchmean'`` | ``'sum'`` | ``'mean'``.
        ``'none'``: no reduction will be applied,
        ``'batchmean'``: the sum of the output will be divided by
            the batchsize,
        ``'sum'``: the output will be summed,
        ``'mean'``: the output will be divided by the number of
            elements in the output.
        Default: ``'batchmean'``
    loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(
        self,
        tau: float = 1.0,
        alpha: float = 1.0,
        beta: float = 1.0,
        reduction: str = 'batchmean',
        loss_weight: float = 1.0,
    ) -> None:
        super(DKDLoss, self).__init__()
        self.tau = tau
        accept_reduction = {'none', 'batchmean', 'sum', 'mean'}
        assert reduction in accept_reduction, \
            f'KLDivergence supports reduction {accept_reduction}, ' \
            f'but gets {reduction}.'
        self.reduction = reduction
        self.alpha = alpha
        self.beta = beta
        self.loss_weight = loss_weight

    def forward(
        self,
        preds_S: torch.Tensor,
        preds_T: torch.Tensor,
        gt_labels: torch.Tensor,
    ) -> torch.Tensor:
        """DKDLoss forward function.

        Args:
            preds_S (torch.Tensor): The student model prediction, shape (N, C).
            preds_T (torch.Tensor): The teacher model prediction, shape (N, C).
            gt_labels (torch.Tensor): The gt label tensor, shape (N, C).

        Return:
            torch.Tensor: The calculated loss value.
        """
        gt_mask = self._get_gt_mask(preds_S, gt_labels)
        tckd_loss = self._get_tckd_loss(preds_S, preds_T, gt_labels, gt_mask)
        nckd_loss = self._get_nckd_loss(preds_S, preds_T, gt_mask)
        loss = self.alpha * tckd_loss + self.beta * nckd_loss
        return self.loss_weight * loss

    def _get_nckd_loss(
        self,
        preds_S: torch.Tensor,
        preds_T: torch.Tensor,
        gt_mask: torch.Tensor,
    ) -> torch.Tensor:
        """Calculate non-target class knowledge distillation."""
        # implementation to mask out gt_mask, faster than index
        s_nckd = F.log_softmax(preds_S / self.tau - 1000.0 * gt_mask, dim=1)
        t_nckd = F.softmax(preds_T / self.tau - 1000.0 * gt_mask, dim=1)
        return self._kl_loss(s_nckd, t_nckd)

    def _get_tckd_loss(
        self,
        preds_S: torch.Tensor,
        preds_T: torch.Tensor,
        gt_labels: torch.Tensor,
        gt_mask: torch.Tensor,
    ) -> torch.Tensor:
        """Calculate target class knowledge distillation."""
        non_gt_mask = self._get_non_gt_mask(preds_S, gt_labels)
        s_tckd = F.softmax(preds_S / self.tau, dim=1)
        t_tckd = F.softmax(preds_T / self.tau, dim=1)
        mask_student = torch.log(self._cat_mask(s_tckd, gt_mask, non_gt_mask))
        mask_teacher = self._cat_mask(t_tckd, gt_mask, non_gt_mask)
        return self._kl_loss(mask_student, mask_teacher)

    def _kl_loss(
        self,
        preds_S: torch.Tensor,
        preds_T: torch.Tensor,
    ) -> torch.Tensor:
        """Calculate the KL Divergence."""
        kl_loss = F.kl_div(
            preds_S, preds_T, size_average=False,
            reduction=self.reduction) * self.tau**2
        return kl_loss

    def _cat_mask(
        self,
        tckd: torch.Tensor,
        gt_mask: torch.Tensor,
        non_gt_mask: torch.Tensor,
    ) -> torch.Tensor:
        """Calculate preds of target (pt) & preds of non-target (pnt)."""
        t1 = (tckd * gt_mask).sum(dim=1, keepdims=True)
        t2 = (tckd * non_gt_mask).sum(dim=1, keepdims=True)
        return torch.cat([t1, t2], dim=1)

    def _get_gt_mask(
        self,
        logits: torch.Tensor,
        target: torch.Tensor,
    ) -> torch.Tensor:
        """Calculate groundtruth mask on logits with target class tensor.

        Args:
            logits (torch.Tensor): The prediction logits with shape (N, C).
            target (torch.Tensor): The gt_label target with shape (N, C).

        Return:
            torch.Tensor: The masked logits.
        """
        target = target.reshape(-1)
        return torch.zeros_like(logits).scatter_(1, target.unsqueeze(1),
                                                 1).bool()

    def _get_non_gt_mask(
        self,
        logits: torch.Tensor,
        target: torch.Tensor,
    ) -> torch.Tensor:
        """Calculate non-groundtruth mask on logits with target class tensor.

        Args:
            logits (torch.Tensor): The prediction logits with shape (N, C).
            target (torch.Tensor): The gt_label target with shape (N, C).

        Return:
            torch.Tensor: The masked logits.
        """
        target = target.reshape(-1)
        return torch.ones_like(logits).scatter_(1, target.unsqueeze(1),
                                                0).bool()



_base_ = ['./spos_mobilenet_supernet_8xb128_in1k.py']

# FIXME: you may replace this with the mutable_cfg searched by yourself
fix_subnet = 'configs/nas/spos/AngleNAS_SHUFFLENETV2_IN1k_2.0.yaml'  # noqa: E501

model = dict(fix_subnet=fix_subnet)

find_unused_parameters = False



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np

from mmrazor.registry import TASK_UTILS
from ..flops_params_counter import get_model_parameters_number
from .base_counter import BaseCounter


class PoolCounter(BaseCounter):
    """FLOPs/params counter for Pooling series."""

    @staticmethod
    def add_count_hook(module, input, output):
        input = input[0]
        module.__flops__ += int(np.prod(input.shape))
        module.__params__ += get_model_parameters_number(module)


@TASK_UTILS.register_module()
class MaxPool1dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class MaxPool2dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class MaxPool3dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AvgPool1dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AvgPool2dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AvgPool3dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AdaptiveMaxPool1dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AdaptiveMaxPool2dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AdaptiveMaxPool3dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AdaptiveAvgPool1dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AdaptiveAvgPool2dCounter(PoolCounter):
    pass


@TASK_UTILS.register_module()
class AdaptiveAvgPool3dCounter(PoolCounter):
    pass



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

from torch import nn

from mmrazor.registry import TASK_UTILS
from .base_recorder import BaseRecorder


@TASK_UTILS.register_module()
class ParameterRecorder(BaseRecorder):
    """Recorder for Pytorch model's parameters.

    Examples:
        >>> from torch import nn
        >>> class ToyModel(nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.toy_conv = nn.Conv2d(1,1,1)
        ...     def forward(self, x):
        ...         return self.toy_conv(x)

        >>> model = ToyModel()
        >>> [ name for name,_ in model.named_parameters() ]
        ['toy_conv.weight', 'toy_conv.bias']

        >>> recorder = ParameterRecorder('toy_conv.weight')
        >>> recorder.initialize(model)

        >>> recorder.data_buffer
        [Parameter containing: tensor([[[[0.3244]]]], requires_grad=True)]
        >>> recorder.get_record_data()
        Parameter containing: tensor([[[[0.3244]]]], requires_grad=True)
    """

    def prepare_from_model(self, model: Optional[nn.Module] = None) -> None:
        """Record the Pytorch model's parameters."""
        assert model is not None, \
            'model can not be None when use ParameterRecorder.'

        founded = False
        for param_name, param in model.named_parameters():
            if param_name == self.source:
                self.data_buffer.append(param)
                founded = True
                break

        assert founded, f'"{self.source}" is not in the model.'

    def reset_data_buffer(self):
        """Clear data in data_buffer.

        Note:
            The data_buffer stores the address of the parameter in memory and
            does not need to be reset.
        """
        pass



_base_ = [
    'mmrazor::_base_/settings/imagenet_bs1024_spos.py',
    'mmrazor::_base_/nas_backbones/spos_shufflenet_supernet.py',
    'mmcls::_base_/default_runtime.py',
]

# model
supernet = dict(
    type='ImageClassifier',
    data_preprocessor=_base_.preprocess_cfg,
    backbone=_base_.nas_backbone,
    neck=dict(type='GlobalAveragePooling'),
    head=dict(
        type='LinearClsHead',
        num_classes=1000,
        in_channels=1024,
        loss=dict(
            type='LabelSmoothLoss',
            num_classes=1000,
            label_smooth_val=0.1,
            mode='original',
            loss_weight=1.0),
        topk=(1, 5)))

model = dict(
    type='mmrazor.SPOS',
    architecture=supernet,
    mutator=dict(type='mmrazor.OneShotModuleMutator'))

find_unused_parameters = True



# Copyright (c) OpenMMLab. All rights reserved.
from .flops_params_counter import (get_model_complexity_info,
                                   params_units_convert)
from .latency_counter import repeat_measure_inference_speed
from .op_counters import *  # noqa: F401,F403

__all__ = [
    'get_model_complexity_info', 'params_units_convert',
    'repeat_measure_inference_speed'
]



_base_ = [
    'mmcls::_base_/datasets/imagenet_bs32.py',
    'mmcls::_base_/schedules/imagenet_bs256.py',
    'mmcls::_base_/default_runtime.py'
]

model = dict(
    _scope_='mmrazor',
    type='SingleTeacherDistill',
    data_preprocessor=dict(
        type='ImgDataPreprocessor',
        # RGB format normalization parameters
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        # convert image from BGR to RGB
        bgr_to_rgb=True),
    architecture=dict(
        cfg_path='mmcls::resnet/resnet18_8xb32_in1k.py', pretrained=False),
    teacher=dict(
        cfg_path='mmcls::resnet/resnet34_8xb32_in1k.py', pretrained=True),
    teacher_ckpt='resnet34_8xb32_in1k_20210831-f257d4e6.pth',
    distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(
            fc=dict(type='ModuleOutputs', source='head.fc')),
        teacher_recorders=dict(
            fc=dict(type='ModuleOutputs', source='head.fc')),
        distill_losses=dict(
            loss_kl=dict(type='KLDivergence', tau=1, loss_weight=5)),
        loss_forward_mappings=dict(
            loss_kl=dict(
                preds_S=dict(from_student=True, recorder='fc'),
                preds_T=dict(from_student=False, recorder='fc')))))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from .dump_subnet_hook import DumpSubnetHook
from .estimate_resources_hook import EstimateResourcesHook

__all__ = ['DumpSubnetHook', 'EstimateResourcesHook']



# Copyright (c) OpenMMLab. All rights reserved.
from pathlib import Path
from typing import Any, Dict, List, Union

FixMutable = Dict[str, Any]
ValidFixMutable = Union[str, Path, FixMutable]

# RANDOM_SUBNET means the subnet sampled by one or more mutators. Usually used
# for supernet training or searching.

# `SingleMutatorRandomSubnet`` sampled by a mutator, its format is a dict, the
# keys of the dict are the group_id in the mutator‘s search groups, and the
# values ​​of the dict are the choices corresponding to all mutables in each
# search group.

# One search group may contains N mutables. More details of search groups see
# docs for :class:`mmrazor.models.mutators.OneShotModuleMutator`.
SingleMutatorRandomSubnet = Dict[int, Any]

# For some more complex algorithms, multiple mutators may be used, and the
# corresponding format will be a list
MultiMutatorsRandomSubnet = List[SingleMutatorRandomSubnet]

SupportRandomSubnet = Union[SingleMutatorRandomSubnet,
                            MultiMutatorsRandomSubnet]



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Callable, Dict

from .path import (Path, PathConcatNode, PathConvNode, PathDepthWiseConvNode,
                   PathLinearNode, PathList, PathNormNode)


def _is_leaf_grad_fn(grad_fn):
    """Determine whether the current node is a leaf node."""
    if type(grad_fn).__name__ == 'AccumulateGrad':
        return True
    return False


def parse_conv(tracer, grad_fn, module2name, param2module, cur_path,
               result_paths, visited, shared_module):
    """Parse the backward of a conv layer.

    Example:
        >>> conv = nn.Conv2d(3, 3, 3)
        >>> pseudo_img = torch.rand(1, 3, 224, 224)
        >>> out = conv(pseudo_img)
        >>> out.grad_fn.next_functions
        ((None, 0), (<AccumulateGrad object at 0x0000020E405CBD88>, 0),
        (<AccumulateGrad object at 0x0000020E405CB588>, 0))
        >>> # op.next_functions[0][0] is None means this ThnnConv2DBackward
        >>> # op has no parents
        >>> # op.next_functions[1][0].variable is the weight of this Conv2d
        >>> # module
        >>> # op.next_functions[2][0].variable is the bias of this Conv2d
        >>> # module
    """
    leaf_grad_fn = grad_fn.next_functions[1][0]
    while not _is_leaf_grad_fn(leaf_grad_fn):
        leaf_grad_fn = leaf_grad_fn.next_functions[0][0]
    variable = leaf_grad_fn.variable
    param_id = id(variable)
    module = param2module[param_id]
    name = module2name[module]
    parent = grad_fn.next_functions[0][0]
    if module.in_channels == module.groups:
        cur_path.append(PathDepthWiseConvNode(name))
    else:
        cur_path.append(PathConvNode(name))
    # If a module is not a shared module and it has been visited during
    # forward, its parent modules must have been traced already.
    # However, a shared module will be visited more than once during
    # forward, so it is still need to be traced even if it has been
    # visited.
    if visited[name] and name not in shared_module:
        result_paths.append(copy.deepcopy(cur_path))
    else:
        visited[name] = True
        tracer.backward_trace(parent, module2name, param2module, cur_path,
                              result_paths, visited, shared_module)
    cur_path.pop(-1)


# todo: support parsing `MultiheadAttention` and user-defined matrix
#  multiplication
def parse_linear(tracer, grad_fn, module2name, param2module, cur_path,
                 result_paths, visited, shared_module):
    """Parse the backward of a conv layer.

    Example:
        >>> fc = nn.Linear(3, 3, bias=True)
        >>> input = torch.rand(3, 3)
        >>> out = fc(input)
        >>> out.grad_fn.next_functions
        ((<AccumulateGrad object at 0x0000020E405F75C8>, 0), (None, 0),
        (<TBackward object at 0x0000020E405F7D48>, 0))
        >>> # op.next_functions[0][0].variable is the bias of this Linear
        >>> # module
        >>> # op.next_functions[1][0] is None means this AddmmBackward op
        >>> # has no parents
        >>> # op.next_functions[2][0] is the TBackward op, and
        >>> # op.next_functions[2][0].next_functions[0][0].variable is
        >>> # the transpose of the weight of this Linear module
    """
    leaf_grad_fn = grad_fn.next_functions[-1][0].next_functions[0][0]
    while not _is_leaf_grad_fn(leaf_grad_fn):
        leaf_grad_fn = leaf_grad_fn.next_functions[0][0]
    variable = leaf_grad_fn.variable
    param_id = id(variable)
    module = param2module[param_id]
    name = module2name[module]
    parent = grad_fn.next_functions[-2][0]

    cur_path.append(PathLinearNode(name))
    # If a module is not a shared module and it has been visited during
    # forward, its parent modules must have been traced already.
    # However, a shared module will be visited more than once during
    # forward, so it is still need to be traced even if it has been
    # visited.
    if visited[name] and name not in shared_module:
        result_paths.append(copy.deepcopy(cur_path))
    else:
        visited[name] = True
        tracer.backward_trace(parent, module2name, param2module, cur_path,
                              result_paths, visited, shared_module)


def parse_cat(tracer, grad_fn, module2name, param2module, cur_path,
              result_paths, visited, shared_module):
    """Parse the backward of a concat operation.

    Example:
        >>> conv = nn.Conv2d(3, 3, 3)
        >>> pseudo_img = torch.rand(1, 3, 224, 224)
        >>> out1 = conv(pseudo_img)
        >>> out2 = conv(pseudo_img)
        >>> out = torch.cat([out1, out2], dim=1)
        >>> out.grad_fn.next_functions
        ((<ThnnConv2DBackward object at 0x0000020E405F24C8>, 0),
        (<ThnnConv2DBackward object at 0x0000020E405F2648>, 0))
        >>> # the length of ``out.grad_fn.next_functions`` is two means
        >>> # ``out`` is obtained by concatenating two tensors
    """
    parents = grad_fn.next_functions
    concat_id = '_'.join([str(id(p)) for p in parents])
    name = f'concat_{concat_id}'
    # If a module is not a shared module and it has been visited during
    # forward, its parent modules must have been traced already.
    # However, a shared module will be visited more than once during
    # forward, so it is still need to be traced even if it has been
    # visited.
    if (name in visited and visited[name] and name not in shared_module):
        pass
    else:
        visited[name] = True
        sub_path_lists = list()
        for i, parent in enumerate(parents):
            sub_path_list = PathList()
            tracer.backward_trace(parent, module2name, param2module, Path(),
                                  sub_path_list, visited, shared_module)
            sub_path_lists.append(sub_path_list)
        cur_path.append(PathConcatNode(name, sub_path_lists))

        result_paths.append(copy.deepcopy(cur_path))
        cur_path.pop(-1)


def parse_norm(tracer, grad_fn, module2name, param2module, cur_path,
               result_paths, visited, shared_module):
    """Parse the backward of a concat operation.

    Example:
        >>> conv = nn.Conv2d(3, 3, 3)
        >>> pseudo_img = torch.rand(1, 3, 224, 224)
        >>> out1 = conv(pseudo_img)
        >>> out2 = conv(pseudo_img)
        >>> out = torch.cat([out1, out2], dim=1)
        >>> out.grad_fn.next_functions
        ((<ThnnConv2DBackward object at 0x0000020E405F24C8>, 0),
        (<ThnnConv2DBackward object at 0x0000020E405F2648>, 0))
        >>> # the length of ``out.grad_fn.next_functions`` is two means
        >>> # ``out`` is obtained by concatenating two tensors
    """
    leaf_grad_fn = grad_fn.next_functions[1][0]
    while not _is_leaf_grad_fn(leaf_grad_fn):
        leaf_grad_fn = leaf_grad_fn.next_functions[0][0]
    variable = leaf_grad_fn.variable
    param_id = id(variable)
    module = param2module[param_id]
    name = module2name[module]
    parent = grad_fn.next_functions[0][0]
    cur_path.append(PathNormNode(name))

    visited[name] = True
    tracer.backward_trace(parent, module2name, param2module, cur_path,
                          result_paths, visited, shared_module)
    cur_path.pop(-1)


DEFAULT_BACKWARD_TRACER: Dict[str, Callable] = {
    'ThnnConv2DBackward': parse_conv,
    'CudnnConvolutionBackward': parse_conv,
    'MkldnnConvolutionBackward': parse_conv,
    'SlowConvDilated2DBackward': parse_conv,
    'ThAddmmBackward': parse_linear,
    'AddmmBackward': parse_linear,
    'MmBackward': parse_linear,
    'CatBackward': parse_cat,
    'ThnnBatchNormBackward': parse_norm,
    'CudnnBatchNormBackward': parse_norm,
    'NativeBatchNormBackward': parse_norm,
    'NativeGroupNormBackward': parse_norm
}



_STAGE_MUTABLE = dict(
    _scope_='mmrazor',
    type='OneShotMutableOP',
    candidates=dict(
        shuffle_3x3=dict(type='ShuffleBlock', kernel_size=3),
        shuffle_5x5=dict(type='ShuffleBlock', kernel_size=5),
        shuffle_7x7=dict(type='ShuffleBlock', kernel_size=7),
        shuffle_xception=dict(type='ShuffleXception')))

arch_setting = [
    # Parameters to build layers. 3 parameters are needed to construct a
    # layer, from left to right: channel, num_blocks, mutable_cfg.
    [64, 4, _STAGE_MUTABLE],
    [160, 4, _STAGE_MUTABLE],
    [320, 8, _STAGE_MUTABLE],
    [640, 4, _STAGE_MUTABLE]
]

nas_backbone = dict(
    _scope_='mmrazor',
    type='SearchableShuffleNetV2',
    widen_factor=1.0,
    arch_setting=arch_setting)



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np

from mmrazor.registry import TASK_UTILS
from .base_counter import BaseCounter


class ConvCounter(BaseCounter):
    """FLOPs/params counter for Conv module series."""

    @staticmethod
    def add_count_hook(module, input, output):
        # Can have multiple inputs, getting the first one
        input = input[0]

        batch_size = input.shape[0]
        output_dims = list(output.shape[2:])

        kernel_dims = list(module.kernel_size)
        in_channels = module.in_channels
        out_channels = module.out_channels
        groups = module.groups

        filters_per_channel = out_channels / groups
        conv_per_position_flops = int(
            np.prod(kernel_dims)) * in_channels * filters_per_channel

        active_elements_count = batch_size * int(np.prod(output_dims))

        overall_conv_flops = conv_per_position_flops * active_elements_count
        overall_params = conv_per_position_flops

        bias_flops = 0
        overall_params = conv_per_position_flops
        if module.bias is not None:
            bias_flops = out_channels * active_elements_count
            overall_params += out_channels

        overall_flops = overall_conv_flops + bias_flops

        module.__flops__ += overall_flops
        module.__params__ += int(overall_params)


@TASK_UTILS.register_module()
class Conv1dCounter(ConvCounter):
    pass


@TASK_UTILS.register_module()
class Conv2dCounter(ConvCounter):
    pass


@TASK_UTILS.register_module()
class Conv3dCounter(ConvCounter):
    pass



# Copyright (c) OpenMMLab. All rights reserved.
from .make_divisible import make_divisible
from .misc import add_prefix
from .optim_wrapper import reinitialize_optim_wrapper_count_status
from .utils import get_module_device, set_requires_grad

__all__ = [
    'add_prefix', 'reinitialize_optim_wrapper_count_status', 'make_divisible',
    'get_module_device', 'set_requires_grad'
]



# Copyright (c) OpenMMLab. All rights reserved.
import logging
import time

import torch
from mmengine.logging import print_log


def repeat_measure_inference_speed(model,
                                   resource_args,
                                   max_iter: int = 100,
                                   log_interval: int = 100,
                                   repeat_num: int = 1) -> float:
    """Repeat speed measure for multi-times to get more precise results."""
    assert repeat_num >= 1

    fps_list = []

    for _ in range(repeat_num):

        fps_list.append(
            measure_inference_speed(model, resource_args, max_iter,
                                    log_interval))

    if repeat_num > 1:
        fps_list_ = [round(fps, 1) for fps in fps_list]
        times_per_img_list = [round(1000 / fps, 1) for fps in fps_list]
        mean_fps_ = sum(fps_list_) / len(fps_list_)
        mean_times_per_img = sum(times_per_img_list) / len(times_per_img_list)
        print_log(
            f'Overall fps: {fps_list_}[{mean_fps_:.1f}] img / s, '
            f'times per image: '
            f'{times_per_img_list}[{mean_times_per_img:.1f}] ms/img',
            logger='current',
            level=logging.DEBUG)
        return mean_times_per_img

    latency = round(1000 / fps_list[0], 1)
    return latency


def measure_inference_speed(model, resource_args, max_iter: int,
                            log_interval: int) -> float:
    # the first several iterations may be very slow so skip them
    num_warmup = 5
    pure_inf_time = 0.0
    fps = 0.0
    data = dict()
    if next(model.parameters()).is_cuda:
        device = 'cuda'
    else:
        raise NotImplementedError('To use cpu to test latency not supported.')
    # benchmark with 100 image and take the average
    for i in range(1, max_iter):
        if device == 'cuda':
            data = torch.rand(resource_args['input_shape']).cuda()
        torch.cuda.synchronize()
        start_time = time.perf_counter()

        with torch.no_grad():
            model(data)

        torch.cuda.synchronize()
        elapsed = time.perf_counter() - start_time

        if i >= num_warmup:
            pure_inf_time += elapsed
            if (i + 1) % log_interval == 0:
                fps = (i + 1 - num_warmup) / pure_inf_time
                print_log(
                    f'Done image [{i + 1:<3}/ {max_iter}], '
                    f'fps: {fps:.1f} img / s, '
                    f'times per image: {1000 / fps:.1f} ms / img',
                    logger='current',
                    level=logging.DEBUG)

        if (i + 1) == max_iter:
            fps = (i + 1 - num_warmup) / pure_inf_time
            print_log(
                f'Overall fps: {fps:.1f} img / s, '
                f'times per image: {1000 / fps:.1f} ms / img',
                logger='current',
                level=logging.DEBUG)
            break

        torch.cuda.empty_cache()

    return fps



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import os
import os.path as osp
import random
import warnings
from typing import Dict, List, Optional, Tuple, Union

import torch
from mmengine import fileio
from mmengine.dist import broadcast_object_list
from mmengine.evaluator import Evaluator
from mmengine.runner import EpochBasedTrainLoop
from mmengine.utils import is_list_of
from torch.utils.data import DataLoader

from mmrazor.models.task_modules.estimators import get_model_complexity_info
from mmrazor.registry import LOOPS
from mmrazor.structures import Candidates, export_fix_subnet, load_fix_subnet
from mmrazor.utils import SupportRandomSubnet
from .utils import crossover


@LOOPS.register_module()
class EvolutionSearchLoop(EpochBasedTrainLoop):
    """Loop for evolution searching.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict): A dataloader object or a dict to
            build a dataloader.
        evaluator (Evaluator or dict or list): Used for computing metrics.
        max_epochs (int): Total searching epochs. Defaults to 20.
        max_keep_ckpts (int): The maximum checkpoints of searcher to keep.
            Defaults to 3.
        resume_from (str, optional): Specify the path of saved .pkl file for
            resuming searching.
        num_candidates (int): The length of candidate pool. Defaults to 50.
        top_k (int): Specify top k candidates based on scores. Defaults to 10.
        num_mutation (int): The number of candidates got by mutation.
            Defaults to 25.
        num_crossover (int): The number of candidates got by crossover.
            Defaults to 25.
        mutate_prob (float): The probability of mutation. Defaults to 0.1.
        flops_range (tuple, optional): flops_range to be used for screening
            candidates.
        spec_modules (list): Used for specify modules need to counter.
            Defaults to list().
        score_key (str): Specify one metric in evaluation results to score
            candidates. Defaults to 'accuracy_top-1'.
        init_candidates (str, optional): The candidates file path, which is
            used to init `self.candidates`. Its format is usually in .yaml
            format. Defaults to None.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[DataLoader, Dict],
                 evaluator: Union[Evaluator, Dict, List],
                 max_epochs: int = 20,
                 max_keep_ckpts: int = 3,
                 resume_from: Optional[str] = None,
                 num_candidates: int = 50,
                 top_k: int = 10,
                 num_mutation: int = 25,
                 num_crossover: int = 25,
                 mutate_prob: float = 0.1,
                 flops_range: Optional[Tuple[float, float]] = (0., 330 * 1e6),
                 spec_modules: List = [],
                 score_key: str = 'accuracy/top1',
                 init_candidates: Optional[str] = None) -> None:
        super().__init__(runner, dataloader, max_epochs)
        if isinstance(evaluator, dict) or is_list_of(evaluator, dict):
            self.evaluator = runner.build_evaluator(evaluator)  # type: ignore
        else:
            self.evaluator = evaluator  # type: ignore
        if hasattr(self.dataloader.dataset, 'metainfo'):
            self.evaluator.dataset_meta = self.dataloader.dataset.metainfo
        else:
            warnings.warn(
                f'Dataset {self.dataloader.dataset.__class__.__name__} has no '
                'metainfo. ``dataset_meta`` in evaluator, metric and '
                'visualizer will be None.')

        self.num_candidates = num_candidates
        self.top_k = top_k
        self.flops_range = flops_range
        self.spec_modules = spec_modules
        self.score_key = score_key
        self.num_mutation = num_mutation
        self.num_crossover = num_crossover
        self.mutate_prob = mutate_prob
        self.max_keep_ckpts = max_keep_ckpts
        self.resume_from = resume_from

        if init_candidates is None:
            self.candidates = Candidates()
        else:
            self.candidates = fileio.load(init_candidates)
            assert isinstance(self.candidates, Candidates), 'please use the \
                correct init candidates file'

        self.top_k_candidates = Candidates()

        if self.runner.distributed:
            self.model = runner.model.module
        else:
            self.model = runner.model

    def run(self) -> None:
        """Launch searching."""
        self.runner.call_hook('before_train')

        if self.resume_from:
            self._resume()

        while self._epoch < self._max_epochs:
            self.run_epoch()
            self._save_searcher_ckpt()

        self._save_best_fix_subnet()

        self.runner.call_hook('after_train')

    def run_epoch(self) -> None:
        """Iterate one epoch.

        Steps:
            1. Sample some new candidates from the supernet.Then Append them
                to the candidates, Thus make its number equal to the specified
                number.
            2. Validate these candidates(step 1) and update their scores.
            3. Pick the top k candidates based on the scores(step 2), which
                will be used in mutation and crossover.
            4. Implement Mutation and crossover, generate better candidates.
        """
        self.sample_candidates()
        self.update_candidates_scores()

        scores_before = self.top_k_candidates.scores
        self.runner.logger.info(f'top k scores before update: '
                                f'{scores_before}')

        self.candidates.extend(self.top_k_candidates)
        self.candidates.sort(key=lambda x: x[1], reverse=True)
        self.top_k_candidates = Candidates(self.candidates[:self.top_k])

        scores_after = self.top_k_candidates.scores
        self.runner.logger.info(f'top k scores after update: '
                                f'{scores_after}')

        mutation_candidates = self.gen_mutation_candidates()
        crossover_candidates = self.gen_crossover_candidates()
        candidates = mutation_candidates + crossover_candidates
        assert len(candidates) <= self.num_candidates, 'Total of mutation and \
            crossover should be no more than the number of candidates.'

        self.candidates = Candidates(candidates)
        self._epoch += 1

    def sample_candidates(self) -> None:
        """Update candidate pool contains specified number of candicates."""
        if self.runner.rank == 0:
            while len(self.candidates) < self.num_candidates:
                candidate = self.model.sample_subnet()
                if self._check_constraints(random_subnet=candidate):
                    self.candidates.append(candidate)
        else:
            self.candidates = Candidates([None] * self.num_candidates)
        # broadcast candidates to val with multi-GPUs.
        broadcast_object_list(self.candidates.data)

    def update_candidates_scores(self) -> None:
        """Validate candicate one by one from the candicate pool, and update
        top-k candicates."""
        for i, candidate in enumerate(self.candidates.subnets):
            self.model.set_subnet(candidate)
            metrics = self._val_candidate()
            score = metrics[self.score_key] \
                if len(metrics) != 0 else 0.
            self.candidates.set_score(i, score)
            self.runner.logger.info(
                f'Epoch:[{self._epoch}/{self._max_epochs}] '
                f'Candidate:[{i + 1}/{self.num_candidates}] '
                f'Score:{score}')

    def gen_mutation_candidates(self) -> List:
        """Generate specified number of mutation candicates."""
        mutation_candidates: List = []
        max_mutate_iters = self.num_mutation * 10
        mutate_iter = 0
        while len(mutation_candidates) < self.num_mutation:
            mutate_iter += 1
            if mutate_iter > max_mutate_iters:
                break

            mutation_candidate = self._mutation()

            if self._check_constraints(random_subnet=mutation_candidate):
                mutation_candidates.append(mutation_candidate)
        return mutation_candidates

    def gen_crossover_candidates(self) -> List:
        """Generate specofied number of crossover candicates."""
        crossover_candidates: List = []
        crossover_iter = 0
        max_crossover_iters = self.num_crossover * 10
        while len(crossover_candidates) < self.num_crossover:
            crossover_iter += 1
            if crossover_iter > max_crossover_iters:
                break

            crossover_candidate = self._crossover()

            if self._check_constraints(random_subnet=crossover_candidate):
                crossover_candidates.append(crossover_candidate)
        return crossover_candidates

    def _mutation(self) -> SupportRandomSubnet:
        """Mutate with the specified mutate_prob."""
        candidate1 = random.choice(self.top_k_candidates.subnets)
        candidate2 = self.model.sample_subnet()
        candidate = crossover(candidate1, candidate2, prob=self.mutate_prob)
        return candidate

    def _crossover(self) -> SupportRandomSubnet:
        """Crossover."""
        candidate1 = random.choice(self.top_k_candidates.subnets)
        candidate2 = random.choice(self.top_k_candidates.subnets)
        candidate = crossover(candidate1, candidate2)
        return candidate

    def _resume(self):
        """Resume searching."""
        if self.runner.rank == 0:
            searcher_resume = fileio.load(self.resume_from)
            for k in searcher_resume.keys():
                setattr(self, k, searcher_resume[k])
            epoch_start = int(searcher_resume['_epoch'])
            self._max_epochs = self._max_epochs - epoch_start
            self.runner.logger.info('#' * 100)
            self.runner.logger.info(f'Resume from epoch: {epoch_start}')
            self.runner.logger.info('#' * 100)

    def _save_best_fix_subnet(self):
        """Save best subnet in searched top-k candidates."""
        if self.runner.rank == 0:
            best_random_subnet = self.top_k_candidates.subnets[0]
            self.model.set_subnet(best_random_subnet)
            best_fix_subnet = export_fix_subnet(self.model)
            save_name = 'best_fix_subnet.yaml'
            fileio.dump(best_fix_subnet,
                        osp.join(self.runner.work_dir, save_name))
            self.runner.logger.info(
                'Search finished and '
                f'{save_name} saved in {self.runner.work_dir}.')

    @torch.no_grad()
    def _val_candidate(self) -> Dict:
        """Run validation."""
        self.runner.model.eval()
        for data_batch in self.dataloader:
            outputs = self.runner.model.val_step(data_batch)
            self.evaluator.process(data_samples=outputs, data_batch=data_batch)
        metrics = self.evaluator.evaluate(len(self.dataloader.dataset))
        return metrics

    def _save_searcher_ckpt(self) -> None:
        """Save searcher ckpt, which is different from common ckpt.

        It mainly contains the candicate pool, the top-k candicates with scores
        and the current epoch.
        """
        if self.runner.rank == 0:
            save_for_resume = dict()
            save_for_resume['_epoch'] = self._epoch
            for k in ['candidates', 'top_k_candidates']:
                save_for_resume[k] = getattr(self, k)
            fileio.dump(
                save_for_resume,
                osp.join(self.runner.work_dir,
                         f'search_epoch_{self._epoch}.pkl'))
            self.runner.logger.info(
                f'Epoch:[{self._epoch}/{self._max_epochs}], top1_score: '
                f'{self.top_k_candidates.scores[0]}')

            if self.max_keep_ckpts > 0:
                cur_ckpt = self._epoch + 1
                redundant_ckpts = range(1, cur_ckpt - self.max_keep_ckpts)
                for _step in redundant_ckpts:
                    ckpt_path = osp.join(self.runner.work_dir,
                                         f'search_epoch_{_step}.pkl')
                    if osp.isfile(ckpt_path):
                        os.remove(ckpt_path)

    def _check_constraints(self, random_subnet: SupportRandomSubnet) -> bool:
        """Check whether is beyond constraints.

        Returns:
            bool: The result of checking.
        """
        if self.flops_range is None:
            return True

        self.model.set_subnet(random_subnet)
        fix_mutable = export_fix_subnet(self.model)
        copied_model = copy.deepcopy(self.model)
        load_fix_subnet(copied_model, fix_mutable)
        flops, _ = get_model_complexity_info(
            copied_model, spec_modules=self.spec_modules)

        if self.flops_range[0] <= flops <= self.flops_range[1]:
            return True
        else:
            return False



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import re
from collections import OrderedDict

from mmengine import ConfigDict
from torch.nn import Conv2d, Linear
from torch.nn.modules import GroupNorm
from torch.nn.modules.batchnorm import _NormBase

from mmrazor.registry import TASK_UTILS
from .parsers import DEFAULT_BACKWARD_TRACER
from .path import Path, PathList

SUPPORT_MODULES = (Conv2d, Linear, _NormBase, GroupNorm)


@TASK_UTILS.register_module()
class BackwardTracer:
    """A topology tracer via backward.

    Args:
        loss_calculator (dict or Callable): Calculate the pseudo loss to trace
            the topology of a model.
    """

    def __init__(self, loss_calculator):
        if isinstance(loss_calculator, (dict, ConfigDict)):
            loss_calculator = TASK_UTILS.build(loss_calculator)

        assert callable(
            loss_calculator
        ), 'loss_calculator should be a dict, ConfigDict or ' \
           'callable object'
        self.loss_calculator = loss_calculator

    @property
    def backward_parser(self):
        """The mapping from the type of a backward op to the corresponding
        parser."""
        return DEFAULT_BACKWARD_TRACER

    def backward_trace(self, grad_fn, module2name, param2module, cur_path,
                       result_paths, visited, shared_module):
        """Trace the topology of all the ``NON_PASS_MODULE``."""
        grad_fn = grad_fn[0] if isinstance(grad_fn, (list, tuple)) else grad_fn

        if grad_fn is not None:
            name = type(grad_fn).__name__
            # In pytorch graph, there may be an additional '0' or '1'
            # (e.g. ThnnConv2DBackward0) after a backward op. Delete the
            # digit numbers to build the corresponding parser.
            name = re.sub(r'[0-1]+', '', name)
            parse_module = self.backward_parser.get(name)

            if parse_module is not None:
                parse_module(self, grad_fn, module2name, param2module,
                             cur_path, result_paths, visited, shared_module)
            else:
                # If the op is AccumulateGrad, parents is (),
                parents = grad_fn.next_functions
                if parents is not None:
                    for parent in parents:
                        self.backward_trace(parent, module2name, param2module,
                                            cur_path, result_paths, visited,
                                            shared_module)
        else:
            result_paths.append(copy.deepcopy(cur_path))

    def _trace_shared_module_hook(self, module, inputs, outputs):
        """Trace shared modules. Modules such as the detection head in
        RetinaNet which are visited more than once during :func:`forward` are
        shared modules.

        Args:
            module (:obj:`torch.nn.Module`): The module to register hook.
            inputs (tuple): The input of the module.
            outputs (tuple): The output of the module.
        """
        module._cnt += 1

    def _build_mappings(self, model):
        """Build the mappings which are used during tracing."""

        module2name = OrderedDict()
        # build a mapping from the identity of a module's params
        # to this module
        param2module = OrderedDict()
        # record the visited module name during trace path
        visited = dict()

        def traverse(module, prefix=''):
            for name, child in module.named_children():
                full_name = f'{prefix}.{name}' if prefix else name

                if isinstance(child, SUPPORT_MODULES):
                    module2name[child] = full_name
                    for param in child.parameters():
                        param2module[id(param)] = child
                    visited[full_name] = False
                else:
                    traverse(child, full_name)

        traverse(model)

        return module2name, param2module, visited

    def _register_share_module_hook(self, model):
        """Record shared modules which will be visited more than once during
        forward such as shared detection head in RetinaNet.

        If a module is not a shared module and it has been visited during
        forward, its parent modules must have been traced already. However, a
        shared module will be visited more than once during forward, so it is
        still need to be traced even if it has been visited.
        """
        self._shared_module_hook_handles = list()
        for module in model.modules():
            if hasattr(module, 'weight'):
                # trace shared modules
                module._cnt = 0
                # the handle is only to remove the corresponding hook later
                handle = module.register_forward_hook(
                    self._trace_shared_module_hook)
                self._shared_module_hook_handles.append(handle)

    def _remove_share_module_hook(self, model):
        """`_trace_shared_module_hook` and `_cnt` are only used to trace the
        shared modules in a model and need to be remove later."""
        for module in model.modules():
            if hasattr(module, 'weight'):
                del module._cnt

        for handle in self._shared_module_hook_handles:
            handle.remove()

        del self._shared_module_hook_handles

    def _set_all_requires_grad(self, model):
        """Set `requires_grad` of a parameter to True to trace the whole
        architecture topology."""
        self._param_requires_grad = dict()
        for param in model.parameters():
            self._param_requires_grad[id(param)] = param.requires_grad
            param.requires_grad = True

    def _restore_requires_grad(self, model):
        """We set requires_grad to True to trace the whole architecture
        topology.

        So it should be reset after that.
        """
        for param in model.parameters():
            param.requires_grad = self._param_requires_grad[id(param)]
        del self._param_requires_grad

    @staticmethod
    def _find_share_modules(model):
        """Find shared modules which will be visited more than once during
        forward such as shared detection head in RetinaNet."""
        share_modules = list()
        for name, module in model.named_modules():
            if hasattr(module, 'weight'):
                if module._cnt > 1:
                    share_modules.append(name)

        return share_modules

    @staticmethod
    def _reset_norm_running_stats(model):
        """As we calculate the pseudo loss during tracing, we need to reset
        states of parameters."""
        for module in model.modules():
            if isinstance(module, _NormBase):
                module.reset_parameters()

    def trace(self, model):
        """Trace trace the architecture topology of the input model."""
        module2name, param2module, visited = self._build_mappings(model)

        # Set requires_grad to True. If the `requires_grad` of a module's
        # weight is False, we can not trace this module by parsing backward.
        self._set_all_requires_grad(model)

        self._register_share_module_hook(model)

        pseudo_loss = self.loss_calculator(model)

        share_modules = self._find_share_modules(model)

        self._remove_share_module_hook(model)
        self._restore_requires_grad(model)

        module_path_list = PathList()

        self.backward_trace(pseudo_loss.grad_fn, module2name, param2module,
                            Path(), module_path_list, visited, share_modules)

        self._reset_norm_running_stats(model)

        return module_path_list



# Copyright (c) OpenMMLab. All rights reserved.
from .datafree_distillation import (DAFLDataFreeDistillation,
                                    DataFreeDistillation)
from .fpn_teacher_distill import FpnTeacherDistill
from .overhaul_feature_distillation import OverhaulFeatureDistillation
from .self_distill import SelfDistill
from .single_teacher_distill import SingleTeacherDistill

__all__ = [
    'SelfDistill', 'SingleTeacherDistill', 'FpnTeacherDistill',
    'DataFreeDistillation', 'DAFLDataFreeDistillation',
    'OverhaulFeatureDistillation'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .byot_connector import BYOTConnector
from .convmodule_connector import ConvModuleConncetor
from .factor_transfer_connectors import Paraphraser, Translator
from .fbkd_connector import FBKDStudentConnector, FBKDTeacherConnector
from .ofd_connector import OFDTeacherConnector
from .torch_connector import TorchFunctionalConnector, TorchNNConnector

__all__ = [
    'ConvModuleConncetor', 'Translator', 'Paraphraser', 'BYOTConnector',
    'FBKDTeacherConnector', 'FBKDStudentConnector', 'TorchFunctionalConnector',
    'TorchNNConnector', 'OFDTeacherConnector'
]



_base_ = [
    '../../../_base_/datasets/mmcls/cifar100_bs16_auto_aug.py',
    'mmcls::_base_/schedules/cifar10_bs128.py',
    'mmcls::_base_/default_runtime.py'
]

optim_wrapper = dict(
    optimizer=dict(type='SGD', lr=0.1, momentum=0.9, weight_decay=0.0005))
param_scheduler = dict(
    type='MultiStepLR', by_epoch=True, milestones=[80, 160, 240], gamma=0.1)
train_cfg = dict(by_epoch=True, max_epochs=250, val_interval=1)

model = dict(
    _scope_='mmrazor',
    type='SelfDistill',
    data_preprocessor=dict(
        type='ImgDataPreprocessor',
        # RGB format normalization parameters
        mean=[129.304, 124.070, 112.434],
        std=[68.170, 65.392, 70.418],
        # convert image from BGR to RGB
        bgr_to_rgb=False),
    architecture=dict(
        type='mmcls.ImageClassifier',
        backbone=dict(
            type='mmcls.ResNet_CIFAR',
            depth=18,
            num_stages=4,
            out_indices=(3, ),
            style='pytorch'),
        neck=dict(type='mmcls.GlobalAveragePooling'),
        head=dict(
            type='mmcls.LinearClsHead',
            num_classes=100,
            in_channels=512,
            loss=dict(type='mmcls.CrossEntropyLoss', loss_weight=1.0))),
    distiller=dict(
        type='BYOTDistiller',
        student_recorders=dict(
            bb_s1=dict(type='ModuleOutputs', source='backbone.layer1.1.relu'),
            bb_s2=dict(type='ModuleOutputs', source='backbone.layer2.1.relu'),
            bb_s3=dict(type='ModuleOutputs', source='backbone.layer3.1.relu')),
        teacher_recorders=dict(
            fc=dict(type='ModuleOutputs', source='head.fc'),
            neck_gap=dict(type='ModuleOutputs', source='neck.gap'),
            gt_labels=dict(type='ModuleInputs', source='head.loss_module')),
        distill_losses=dict(
            loss_fet_1=dict(
                type='L2Loss', normalize=False, loss_weight=0.03, dist=True),
            loss_label_1=dict(type='mmcls.CrossEntropyLoss', loss_weight=0.7),
            loss_softl_1=dict(type='KLDivergence', tau=3, loss_weight=0.3),
            loss_fet_2=dict(
                type='L2Loss', normalize=False, loss_weight=0.03, dist=True),
            loss_label_2=dict(type='mmcls.CrossEntropyLoss', loss_weight=0.7),
            loss_softl_2=dict(type='KLDivergence', tau=3, loss_weight=0.3),
            loss_fet_3=dict(
                type='L2Loss', normalize=False, loss_weight=0., dist=True),
            loss_label_3=dict(type='mmcls.CrossEntropyLoss', loss_weight=0.7),
            loss_softl_3=dict(type='KLDivergence', tau=3, loss_weight=0.3)),
        connectors=dict(
            loss_s1_sfeat=dict(
                type='BYOTConnector',
                in_channel=64,
                out_channel=512,
                expansion=1,
                kernel_size=3,
                stride=2,
                num_classes=100),
            loss_s2_sfeat=dict(
                type='BYOTConnector',
                in_channel=128,
                out_channel=512,
                expansion=1,
                kernel_size=3,
                stride=2,
                num_classes=100),
            loss_s3_sfeat=dict(
                type='BYOTConnector',
                in_channel=256,
                out_channel=512,
                expansion=1,
                kernel_size=3,
                stride=2,
                num_classes=100)),
        loss_forward_mappings=dict(
            loss_fet_1=dict(
                s_feature=dict(
                    recorder='bb_s1',
                    from_student=True,
                    connector='loss_s1_sfeat',
                    connector_idx=0),
                t_feature=dict(recorder='neck_gap', from_student=False)),
            loss_label_1=dict(
                cls_score=dict(
                    recorder='bb_s1',
                    from_student=True,
                    connector='loss_s1_sfeat',
                    connector_idx=1),
                label=dict(
                    recorder='gt_labels', from_student=False, data_idx=1)),
            loss_softl_1=dict(
                preds_S=dict(
                    recorder='bb_s1',
                    from_student=True,
                    connector='loss_s1_sfeat',
                    connector_idx=1),
                preds_T=dict(recorder='fc', from_student=False)),
            loss_fet_2=dict(
                s_feature=dict(
                    recorder='bb_s2',
                    from_student=True,
                    connector='loss_s2_sfeat',
                    connector_idx=0),
                t_feature=dict(recorder='neck_gap', from_student=False)),
            loss_label_2=dict(
                cls_score=dict(
                    recorder='bb_s2',
                    from_student=True,
                    connector='loss_s2_sfeat',
                    connector_idx=1),
                label=dict(
                    recorder='gt_labels', from_student=False, data_idx=1)),
            loss_softl_2=dict(
                preds_S=dict(
                    recorder='bb_s2',
                    from_student=True,
                    connector='loss_s2_sfeat',
                    connector_idx=1),
                preds_T=dict(recorder='fc', from_student=False)),
            loss_fet_3=dict(
                s_feature=dict(
                    recorder='bb_s3',
                    from_student=True,
                    connector='loss_s3_sfeat',
                    connector_idx=0),
                t_feature=dict(recorder='neck_gap', from_student=False)),
            loss_label_3=dict(
                cls_score=dict(
                    recorder='bb_s3',
                    from_student=True,
                    connector='loss_s3_sfeat',
                    connector_idx=1),
                label=dict(
                    recorder='gt_labels', from_student=False, data_idx=1)),
            loss_softl_3=dict(
                preds_S=dict(
                    recorder='bb_s3',
                    from_student=True,
                    connector='loss_s3_sfeat',
                    connector_idx=1),
                preds_T=dict(recorder='fc', from_student=False)))))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SelfDistillValLoop')



model = dict(
    _scope_='mmcls',
    type='ImageClassifier',
    backbone=dict(
        _scope_='mmrazor',
        type='WideResNet',
        depth=16,
        num_stages=3,
        widen_factor=2,
    ),
    neck=dict(type='GlobalAveragePooling'),
    head=dict(
        type='LinearClsHead',
        num_classes=10,
        in_channels=128,
        loss=dict(type='CrossEntropyLoss', loss_weight=1.0),
        topk=(1, 5),
    ))

find_unused_parameters = True



# Copyright (c) OpenMMLab. All rights reserved.
import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Set, Tuple

import torch
from mmengine import print_log
from torch import Tensor, nn
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.mutables.base_mutable import BaseMutable


class DynamicMixin(ABC):
    """Base class for dynamic OP. A dynamic OP usually consists of a normal
    static OP and mutables, where mutables are used to control the searchable
    (mutable) part of the dynamic OP.

    Note:
        When the dynamic OP has just been initialized, its forward propagation
        logic should be the same as the corresponding static OP. Only after
        the searchable part accepts the specific mutable through the
        corresponding interface does the part really become dynamic.

    Note:
        All subclass should implement ``to_static_op`` and
        ``static_op_factory`` APIs.

    Args:
        accepted_mutables (set): The string set of all accepted mutables.
    """
    accepted_mutable_attrs: Set[str] = set()
    attr_mappings: Dict[str, str] = dict()

    @abstractmethod
    def register_mutable_attr(self, attr: str, mutable: BaseMutable):
        pass

    def get_mutable_attr(self, attr: str) -> BaseMutable:

        self.check_mutable_attr_valid(attr)
        if attr in self.attr_mappings:
            attr_map = self.attr_mappings[attr]
            return getattr(self.mutable_attrs, attr_map, None)  # type:ignore
        else:
            return getattr(self.mutable_attrs, attr, None)  # type:ignore

    @classmethod
    @abstractmethod
    def convert_from(cls, module):
        """Convert an instance of Pytorch module to a new instance of Dynamic
        module."""

    @property
    @abstractmethod
    def static_op_factory(self):
        """Corresponding Pytorch OP."""

    @abstractmethod
    def to_static_op(self) -> nn.Module:
        """Convert dynamic OP to static OP.

        Note:
            The forward result for the same input between dynamic OP and its
            corresponding static OP must be same.

        Returns:
            nn.Module: Corresponding static OP.
        """

    def check_if_mutables_fixed(self) -> None:
        """Check if all mutables are fixed.

        Raises:
            RuntimeError: Error if a existing mutable is not fixed.
        """

        def check_fixed(mutable: Optional[BaseMutable]) -> None:
            if mutable is not None and not mutable.is_fixed:
                raise RuntimeError(f'Mutable {type(mutable)} is not fixed.')

        for mutable in self.mutable_attrs.values():  # type: ignore
            check_fixed(mutable)

    def check_mutable_attr_valid(self, attr):
        assert attr in self.attr_mappings or \
                    attr in self.accepted_mutable_attrs

    @staticmethod
    def get_current_choice(mutable: BaseMutable) -> Any:
        """Get current choice of given mutable.

        Args:
            mutable (BaseMutable): Given mutable.

        Raises:
            RuntimeError: Error if `current_choice` is None.

        Returns:
            Any: Current choice of given mutable.
        """
        current_choice = mutable.current_choice
        if current_choice is None:
            raise RuntimeError(f'current choice of mutable {type(mutable)} '
                               'can not be None at runtime')

        return current_choice


class DynamicChannelMixin(DynamicMixin):
    """Base class for dynamic OP with mutable channels.

    Note:
        All subclass should implement ``mutable_in_channels`` and
        ``mutable_out_channels`` APIs.
    """

    @staticmethod
    def check_mutable_channels(mutable_channels: BaseMutable) -> None:
        """Check if mutable has `currnet_mask` attribute.

        Args:
            mutable_channels (BaseMutable): Mutable to be checked.

        Raises:
            ValueError: Error if mutable does not have `current_mask`
                attribute.
        """
        if not hasattr(mutable_channels, 'current_mask'):
            raise ValueError(
                'channel mutable must have attribute `current_mask`')


class DynamicBatchNormMixin(DynamicChannelMixin):
    """A mixin class for Pytorch BatchNorm, which can mutate
    ``num_features``."""
    accepted_mutable_attrs: Set[str] = {'num_features'}
    attr_mappings: Dict[str, str] = {
        'in_channels': 'num_features',
        'out_channels': 'num_features',
    }

    def register_mutable_attr(self, attr, mutable):
        self.check_mutable_attr_valid(attr)
        if attr in self.attr_mappings:
            attr_map = self.attr_mappings[attr]
            assert attr_map in self.accepted_mutable_attrs
            if attr_map in self.mutable_attrs:
                print_log(
                    f'{attr_map}({attr}) is already in `mutable_attrs`',
                    level=logging.WARNING)
            else:
                self._register_mutable_attr(attr_map, mutable)
        elif attr in self.accepted_mutable_attrs:
            self._register_mutable_attr(attr, mutable)
        else:
            raise NotImplementedError

    def _register_mutable_attr(self, attr, mutable):

        if attr == 'num_features':
            self._register_mutable_num_features(mutable)
        else:
            raise NotImplementedError

    def _register_mutable_num_features(
            self: _BatchNorm, mutable_num_features: BaseMutable) -> None:
        """Mutate ``num_features`` with given mutable.

        Args:
            mutable_num_features (BaseMutable): Mutable for controlling
                ``num_features``.

        Raises:
            RuntimeError: Error if both ``affine`` and
                ``tracking_running_stats`` are False.
            ValueError: Error if size of mask if not same as ``num_features``.
        """
        if not self.affine and not self.track_running_stats:
            raise RuntimeError(
                'num_features can not be mutated if both `affine` and '
                '`tracking_running_stats` are False')

        self.check_mutable_channels(mutable_num_features)
        mask_size = mutable_num_features.current_mask.size(0)
        if mask_size != self.num_features:
            raise ValueError(
                f'Expect mask size of mutable to be {self.num_features} as '
                f'`num_features`, but got: {mask_size}.')

        self.mutable_attrs['num_features'] = mutable_num_features

    def _get_num_features_mask(self: _BatchNorm) -> Optional[torch.Tensor]:
        """Get mask of ``num_features``"""
        if self.affine:
            refer_tensor = self.weight
        elif self.track_running_stats:
            refer_tensor = self.running_mean
        else:
            return None

        if 'num_features' in self.mutable_attrs:
            out_mask = self.mutable_attrs['num_features'].current_mask.to(
                refer_tensor.device)
        else:
            out_mask = torch.ones_like(refer_tensor).bool()

        return out_mask

    def get_dynamic_params(
        self: _BatchNorm
    ) -> Tuple[Optional[Tensor], Optional[Tensor], Optional[Tensor],
               Optional[Tensor]]:
        """Get dynamic parameters that will be used in forward process.

        Returns:
            Tuple[Optional[Tensor], Optional[Tensor], Optional[Tensor],
                Optional[Tensor]]: Sliced running_mean, running_var, weight and
                bias.
        """
        out_mask = self._get_num_features_mask()

        if self.affine:
            weight = self.weight[out_mask]
            bias = self.bias[out_mask]
        else:
            weight, bias = self.weight, self.bias

        if self.track_running_stats:
            running_mean = self.running_mean[out_mask] \
                if not self.training or self.track_running_stats else None
            running_var = self.running_var[out_mask] \
                if not self.training or self.track_running_stats else None
        else:
            running_mean, running_var = self.running_mean, self.running_var

        return running_mean, running_var, weight, bias

    def to_static_op(self: _BatchNorm) -> nn.Module:
        """Convert dynamic BatchNormxd to :obj:`torch.nn.BatchNormxd`.

        Returns:
            torch.nn.BatchNormxd: :obj:`torch.nn.BatchNormxd` with sliced
                parameters.
        """
        self.check_if_mutables_fixed()

        running_mean, running_var, weight, bias = self.get_dynamic_params()
        if 'num_features' in self.mutable_attrs:
            num_features = self.mutable_attrs['num_features'].current_mask.sum(
            ).item()
        else:
            num_features = self.num_features

        static_bn = self.static_op_factory(
            num_features=num_features,
            eps=self.eps,
            momentum=self.momentum,
            affine=self.affine,
            track_running_stats=self.track_running_stats)

        if running_mean is not None:
            static_bn.running_mean.copy_(running_mean)
        if running_var is not None:
            static_bn.running_var.copy_(running_var)
        if weight is not None:
            static_bn.weight = nn.Parameter(weight)
        if bias is not None:
            static_bn.bias = nn.Parameter(bias)

        return static_bn


class DynamicLinearMixin(DynamicChannelMixin):
    """A mixin class for Pytorch Linear, which can mutate ``in_features`` and
    ``out_features``."""

    accepted_mutable_attrs: Set[str] = {'in_features', 'out_features'}
    attr_mappings: Dict[str, str] = {
        'in_channels': 'in_features',
        'out_channels': 'out_features',
    }

    def register_mutable_attr(self, attr, mutable):
        self.check_mutable_attr_valid(attr)
        if attr in self.attr_mappings:
            attr_map = self.attr_mappings[attr]
            assert attr_map in self.accepted_mutable_attrs
            if attr_map in self.mutable_attrs:
                print_log(
                    f'{attr_map}({attr}) is already in `mutable_attrs`',
                    level=logging.WARNING)
            else:
                self._register_mutable_attr(attr_map, mutable)
        elif attr in self.accepted_mutable_attrs:
            self._register_mutable_attr(attr, mutable)
        else:
            raise NotImplementedError

    def _register_mutable_attr(self, attr, mutable):

        if attr == 'in_features':
            self._register_mutable_in_features(mutable)
        elif attr == 'out_features':
            self._register_mutable_out_features(mutable)
        else:
            raise NotImplementedError

    def _register_mutable_in_features(
            self: nn.Linear, mutable_in_features: BaseMutable) -> None:
        """Mutate ``in_features`` with given mutable.

        Args:
            mutable_in_features (BaseMutable): Mutable for controlling
                ``in_features``.

        Raises:
            ValueError: Error if size of mask if not same as ``in_features``.
        """
        self.check_mutable_channels(mutable_in_features)
        mask_size = mutable_in_features.current_mask.size(0)
        if mask_size != self.in_features:
            raise ValueError(
                f'Expect mask size of mutable to be {self.in_features} as '
                f'`in_features`, but got: {mask_size}.')

        self.mutable_attrs['in_features'] = mutable_in_features

    def _register_mutable_out_features(
            self: nn.Linear, mutable_out_features: BaseMutable) -> None:
        """Mutate ``out_features`` with given mutable.

        Args:
            mutable_out_features (BaseMutable): Mutable for controlling
                ``out_features``.

        Raises:
            ValueError: Error if size of mask if not same as ``out_features``.
        """
        self.check_mutable_channels(mutable_out_features)
        mask_size = mutable_out_features.current_mask.size(0)
        if mask_size != self.out_features:
            raise ValueError(
                f'Expect mask size of mutable to be {self.out_features} as '
                f'`in_features`, but got: {mask_size}.')

        self.mutable_attrs['out_features'] = mutable_out_features

    def get_dynamic_params(self: nn.Linear) -> Tuple[Tensor, Optional[Tensor]]:
        """Get dynamic parameters that will be used in forward process.

        Returns:
            Tuple[Tensor, Optional[Tensor]]: Sliced weight and bias.
        """
        if 'in_features' not in self.mutable_attrs and \
                'out_features' not in self.mutable_attrs:
            return self.weight, self.bias

        if 'in_features' in self.mutable_attrs:
            in_mask = self.mutable_attrs['in_features'].current_mask.to(
                self.weight.device)
        else:
            in_mask = torch.ones(self.weight.size(1)).bool().to(
                self.weight.device)
        if 'out_features' in self.mutable_attrs:

            out_mask = self.mutable_attrs['out_features'].current_mask.to(
                self.weight.device)
        else:
            out_mask = torch.ones(self.weight.size(0)).bool().to(
                self.weight.device)

        weight = self.weight[out_mask][:, in_mask]
        bias = self.bias[out_mask] if self.bias is not None else None

        return weight, bias

    def to_static_op(self: nn.Linear) -> nn.Module:
        """Convert to :obj:`torch.nn.Linear`.

        Returns:
            nn.Linear: :obj:`torch.nn.Linear` with sliced parameters.
        """
        self.check_if_mutables_fixed()

        weight, bias = self.get_dynamic_params()
        out_features = weight.size(0)
        in_features = weight.size(1)

        static_linear = self.static_op_factory(
            in_features=in_features,
            out_features=out_features,
            bias=True if bias is not None else False)

        static_linear.weight = nn.Parameter(weight)
        if bias is not None:
            static_linear.bias = nn.Parameter(bias)

        return static_linear



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS
from .base_connector import BaseConnector

FUNCTION_LIST = [
    'adaptive_avg_pool2d',
    'adaptive_max_pool2d',
    'avg_pool2d',
    'dropout',
    'dropout2d',
    'max_pool2d',
    'normalize',
    'relu',
    'softmax',
    'interpolate',
]


@MODELS.register_module()
class TorchFunctionalConnector(BaseConnector):
    """TorchFunctionalConnector: Call function in torch.nn.functional
    to process input data

    usage:
        tensor1 = torch.rand(3,3,16,16)
        pool_connector = TorchFunctionalConnector(
                            function_name='avg_pool2d',
                            func_args=dict(kernel_size=4),
                        )
        tensor2 = pool_connector.forward_train(tensor1)
        tensor2.size()
        # torch.Size([3, 3, 4, 4])

        which is equal to torch.nn.functional.avg_pool2d(kernel_size=4)
    Args:
        function_name (str, optional): function. Defaults to None.
        func_args (dict): args parsed to function. Defaults to {}.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self,
                 function_name: Optional[str] = None,
                 func_args: Dict = {},
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg)
        assert function_name is not None, 'Arg `function_name` cannot be None'
        if function_name not in FUNCTION_LIST:
            raise ValueError(
                ' Arg `function_name` are not available, See this list',
                FUNCTION_LIST)
        self.func = getattr(F, function_name)
        self.func_args = func_args

    def forward_train(self, x: torch.Tensor) -> torch.Tensor:
        """Frorward function for training.

        Args:
            x (torch.Tensor): Input features.
        """
        x = self.func(x, **self.func_args)
        return x


MODULE_LIST = [
    'AdaptiveAvgPool2d',
    'AdaptiveMaxPool2d',
    'AvgPool2d',
    'BatchNorm2d',
    'Conv2d',
    'Dropout',
    'Dropout2d',
    'Linear',
    'MaxPool2d',
    'ReLU',
    'Softmax',
]


@MODELS.register_module()
class TorchNNConnector(BaseConnector):
    """TorchNNConnector: create nn.module in torch.nn to process input data

    usage:
        tensor1 = torch.rand(3,3,16,16)
        pool_connector = TorchNNConnector(
                            module_name='AvgPool2d',
                            module_args=dict(kernel_size=4),
                        )
        tensor2 = pool_connector.forward_train(tensor1)
        tensor2.size()
        # torch.Size([3, 3, 4, 4])

        which is equal to torch.nn.AvgPool2d(kernel_size=4)
    Args:
        module_name (str, optional):
            module name. Defaults to None.
            possible_values:['AvgPool2d',
                            'Dropout2d',
                            'AdaptiveAvgPool2d',
                            'AdaptiveMaxPool2d',
                            'ReLU',
                            'Softmax',
                            'BatchNorm2d',
                            'Linear',]
        module_args (dict):
            args parsed to nn.Module().__init__(). Defaults to {}.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self,
                 module_name: Optional[str] = None,
                 module_args: Dict = {},
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg)
        assert module_name is not None, 'Arg `module_name` cannot be None'
        if module_name not in MODULE_LIST:
            raise ValueError(
                ' Arg `module_name` are not available, See this list',
                MODULE_LIST)
        self.func = getattr(nn, module_name)(**module_args)

    def forward_train(self, x: torch.Tensor) -> torch.Tensor:
        """Frorward function for training.

        Args:
            x (torch.Tensor): Input features.
        """
        x = self.func(x)
        return x



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS


@MODELS.register_module()
class FTLoss(nn.Module):
    """Paraphrasing Complex Network: Network Compression via Factor Transfer,
    NeurIPS 2018.

    https://arxiv.org/pdf/1802.04977.pdf

    Args:
        loss_weight (float, optional): loss weight. Defaults to 1.0.
    """

    def __init__(self, loss_weight: float = 1.0) -> None:
        super(FTLoss, self).__init__()
        self.criterion = nn.L1Loss()
        self.loss_weight = loss_weight

    def forward_train(self, s_feature: torch.Tensor,
                      t_feature: torch.Tensor) -> torch.Tensor:
        """loss computation func."""
        loss = self.criterion(self.factor(s_feature), self.factor(t_feature))
        return loss

    def forward(self, s_feature: torch.Tensor,
                t_feature: torch.Tensor) -> torch.Tensor:
        """the forward func."""
        return self.loss_weight * self.forward_train(s_feature, t_feature)

    def factor(self, x: torch.Tensor) -> torch.Tensor:
        """compute factor."""
        return F.normalize(x.view(x.size(0), -1))



# Copyright (c) OpenMMLab. All rights reserved.
from .base import BaseAlgorithm
from .distill import (DAFLDataFreeDistillation, DataFreeDistillation,
                      FpnTeacherDistill, OverhaulFeatureDistillation,
                      SelfDistill, SingleTeacherDistill)
from .nas import SPOS, AutoSlim, AutoSlimDDP, Darts, DartsDDP
from .pruning import SlimmableNetwork, SlimmableNetworkDDP

__all__ = [
    'SingleTeacherDistill', 'BaseAlgorithm', 'FpnTeacherDistill', 'SPOS',
    'SlimmableNetwork', 'SlimmableNetworkDDP', 'AutoSlim', 'AutoSlimDDP',
    'Darts', 'DartsDDP', 'SelfDistill', 'DataFreeDistillation',
    'DAFLDataFreeDistillation', 'OverhaulFeatureDistillation'
]



_STAGE_MUTABLE = dict(
    _scope_='mmrazor',
    type='OneShotMutableOP',
    candidates=dict(
        mb_k3e3=dict(
            type='MBBlock',
            kernel_size=3,
            expand_ratio=3,
            act_cfg=dict(type='ReLU6')),
        mb_k5e3=dict(
            type='MBBlock',
            kernel_size=5,
            expand_ratio=3,
            act_cfg=dict(type='ReLU6')),
        mb_k7e3=dict(
            type='MBBlock',
            kernel_size=7,
            expand_ratio=3,
            act_cfg=dict(type='ReLU6')),
        mb_k3e6=dict(
            type='MBBlock',
            kernel_size=3,
            expand_ratio=6,
            act_cfg=dict(type='ReLU6')),
        mb_k5e6=dict(
            type='MBBlock',
            kernel_size=5,
            expand_ratio=6,
            act_cfg=dict(type='ReLU6')),
        mb_k7e6=dict(
            type='MBBlock',
            kernel_size=7,
            expand_ratio=6,
            act_cfg=dict(type='ReLU6')),
        identity=dict(type='Identity')))

_FIRST_MUTABLE = dict(
    _scope_='mmrazor',
    type='OneShotMutableOP',
    candidates=dict(
        mb_k3e1=dict(
            type='MBBlock',
            kernel_size=3,
            expand_ratio=1,
            act_cfg=dict(type='ReLU6'))))

arch_setting = [
    # Parameters to build layers. 3 parameters are needed to construct a
    # layer, from left to right: channel, num_blocks, mutable_cfg.
    [24, 1, 1, _FIRST_MUTABLE],
    [32, 4, 2, _STAGE_MUTABLE],
    [56, 4, 2, _STAGE_MUTABLE],
    [112, 4, 2, _STAGE_MUTABLE],
    [128, 4, 1, _STAGE_MUTABLE],
    [256, 4, 2, _STAGE_MUTABLE],
    [432, 1, 1, _STAGE_MUTABLE]
]

nas_backbone = dict(
    _scope_='mmrazor',
    type='SearchableMobileNet',
    first_channels=40,
    last_channels=1728,
    widen_factor=1.0,
    arch_setting=arch_setting)



# Copyright (c) OpenMMLab. All rights reserved.
from .activation_layer_counter import (ELUCounter, LeakyReLUCounter,
                                       PReLUCounter, ReLU6Counter, ReLUCounter)
from .base_counter import BaseCounter
from .conv_layer_counter import Conv1dCounter, Conv2dCounter, Conv3dCounter
from .deconv_layer_counter import ConvTranspose2dCounter
from .linear_layer_counter import LinearCounter
from .norm_layer_counter import (BatchNorm1dCounter, BatchNorm2dCounter,
                                 BatchNorm3dCounter, GroupNormCounter,
                                 InstanceNorm1dCounter, InstanceNorm2dCounter,
                                 InstanceNorm3dCounter, LayerNormCounter)
from .pooling_layer_counter import *  # noqa: F403, F405, F401
from .upsample_layer_counter import UpsampleCounter

__all__ = [
    'ReLUCounter', 'PReLUCounter', 'ELUCounter', 'LeakyReLUCounter',
    'ReLU6Counter', 'BatchNorm1dCounter', 'BatchNorm2dCounter',
    'BatchNorm3dCounter', 'Conv1dCounter', 'Conv2dCounter', 'Conv3dCounter',
    'ConvTranspose2dCounter', 'UpsampleCounter', 'LinearCounter',
    'GroupNormCounter', 'InstanceNorm1dCounter', 'InstanceNorm2dCounter',
    'InstanceNorm3dCounter', 'LayerNormCounter', 'BaseCounter'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .backbones import *  # noqa: F401,F403
from .connectors import *  # noqa: F401,F403
from .dynamic_ops import *  # noqa: F401,F403
from .generators import *  # noqa: F401,F403
from .heads import *  # noqa: F401,F403
from .ops import *  # noqa: F401,F403



# Policy for CIFAR, refer to
# https://github.com/DeepVoltaire/AutoAugment/blame/master/autoaugment.py
policy_cifar = [
    # Group 1
    [
        dict(type='Invert', prob=0.1),
        dict(type='Contrast', magnitude=0.5, prob=0.2)
    ],
    [
        dict(type='Rotate', angle=10., prob=0.7),
        dict(type='Translate', magnitude=150 / 331, prob=0.3)
    ],
    [
        dict(type='Sharpness', magnitude=0.9, prob=0.8),
        dict(type='Sharpness', magnitude=0.3, prob=0.9)
    ],
    [
        dict(
            type='Shear',
            magnitude=0.3 / 9 * 8,
            direction='vertical',
            prob=0.5),
        dict(
            type='Translate',
            magnitude=150 / 331,
            direction='vertical',
            prob=0.3)
    ],
    [dict(type='AutoContrast', prob=0.5),
     dict(type='Equalize', prob=0.9)],
    # Group 2
    [
        dict(
            type='Shear',
            magnitude=0.3 / 9 * 7,
            direction='vertical',
            prob=0.2),
        dict(type='Posterize', bits=5, prob=0.3)
    ],
    [
        dict(type='ColorTransform', magnitude=0.3, prob=0.4),
        dict(type='Brightness', magnitude=0.7, prob=0.7)
    ],
    [
        dict(type='Sharpness', magnitude=1.0, prob=0.3),
        dict(type='Brightness', magnitude=1.0, prob=0.7)
    ],
    [dict(type='Equalize', prob=0.6),
     dict(type='Equalize', prob=0.5)],
    [
        dict(type='Contrast', magnitude=0.6, prob=0.6),
        dict(type='Sharpness', magnitude=0.4, prob=0.8),
    ],
    # Group 3
    [
        dict(type='ColorTransform', magnitude=0.6, prob=0.7),
        dict(type='Translate', magnitude=150 / 331 / 9 * 7, prob=0.5)
    ],
    [dict(type='Equalize', prob=0.3),
     dict(type='AutoContrast', prob=0.4)],
    [
        dict(
            type='Translate',
            magnitude=150 / 331 / 9 * 2,
            direction='vertical',
            prob=0.4),
        dict(type='Sharpness', magnitude=0.5, prob=0.2)
    ],
    [
        dict(type='Brightness', magnitude=0.5, prob=0.9),
        dict(type='ColorTransform', magnitude=0.7, prob=0.2),
    ],
    [
        dict(type='Solarize', thr=256 / 9 * 7, prob=0.5),
        dict(type='Invert', prob=0.0),
    ],
    # Group 4
    [dict(type='Equalize', prob=0.2),
     dict(type='AutoContrast', prob=0.6)],
    [dict(type='Equalize', prob=0.2),
     dict(type='Equalize', prob=0.6)],
    [
        dict(type='ColorTransform', magnitude=0.9, prob=0.9),
        dict(type='Equalize', prob=0.6)
    ],
    [
        dict(type='AutoContrast', prob=0.8),
        dict(type='Solarize', thr=256 / 9 * 1, prob=0.2),
    ],
    [
        dict(type='Brightness', magnitude=0.3, prob=0.1),
        dict(type='ColorTransform', magnitude=0.0, prob=0.7)
    ],
    # Group 5
    [
        dict(type='Solarize', thr=256 / 9 * 4, prob=0.4),
        dict(type='AutoContrast', prob=0.9)
    ],
    [
        dict(
            type='Translate',
            magnitude=150 / 331,
            direction='vertical',
            prob=0.9),
        dict(
            type='Translate',
            magnitude=150 / 331,
            direction='vertical',
            prob=0.7)
    ],
    [
        dict(type='AutoContrast', prob=0.9),
        dict(type='Solarize', thr=256 / 9 * 6, prob=0.8)
    ],
    [dict(type='Equalize', prob=0.8),
     dict(type='Invert', prob=0.1)],
    [
        dict(
            type='Translate',
            magnitude=150 / 331,
            direction='vertical',
            prob=0.7),
        dict(type='AutoContrast', prob=0.9)
    ]
]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
from pathlib import Path

import torch


def parse_args():
    parser = argparse.ArgumentParser(
        description='Process a checkpoint to be published')
    parser.add_argument('checkpoint', help='input checkpoint filename')
    parser.add_argument(
        '--inplace', action='store_true', help='replace origin ckpt')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    checkpoint = torch.load(args.checkpoint, map_location='cpu')
    new_state_dict = dict()

    for key, value in checkpoint['state_dict'].items():
        if key.startswith('architecture.model.distiller.teacher'):
            new_key = key.replace('architecture.model.distiller.teacher',
                                  'architecture.teacher')
        elif key.startswith('architecture.model'):
            new_key = key.replace('architecture.model', 'architecture')
        else:
            new_key = key

        new_state_dict[new_key] = value

    checkpoint['state_dict'] = new_state_dict

    if args.inplace:
        torch.save(checkpoint, args.checkpoint)
    else:
        ckpt_path = Path(args.checkpoint)
        ckpt_name = ckpt_path.stem
        ckpt_dir = ckpt_path.parent
        new_ckpt_path = ckpt_dir / f'{ckpt_name}_latest.pth'
        torch.save(checkpoint, new_ckpt_path)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
from mmrazor.registry import TASK_UTILS
from ..flops_params_counter import get_model_parameters_number
from .base_counter import BaseCounter


@TASK_UTILS.register_module()
class UpsampleCounter(BaseCounter):
    """FLOPs/params counter for Upsample function."""

    @staticmethod
    def add_count_hook(module, input, output):
        output_size = output[0]
        batch_size = output_size.shape[0]
        output_elements_count = batch_size
        for val in output_size.shape[1:]:
            output_elements_count *= val
        module.__flops__ += int(output_elements_count)
        module.__params__ += get_model_parameters_number(module)



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Tuple

import torch.nn

from mmrazor.registry import TASK_UTILS


@TASK_UTILS.register_module()
class BaseEstimator(metaclass=ABCMeta):
    """The base class of Estimator, used for estimating model infos.

    Args:
        default_shape (tuple): Input data's default shape, for calculating
            resources consume. Defaults to (1, 3, 224, 224).
        units (str): Resource units. Defaults to 'M'.
        disabled_counters (list): List of disabled spec op counters.
            Defaults to None.
        as_strings (bool): Output FLOPs and params counts in a string
            form. Default to False.
        measure_inference (bool): whether to measure infer speed or not.
            Default to False.
    """

    def __init__(self,
                 default_shape: Tuple = (1, 3, 224, 224),
                 units: str = 'M',
                 disabled_counters: List[str] = None,
                 as_strings: bool = False,
                 measure_inference: bool = False):
        assert len(default_shape) in [3, 4, 5], \
            f'Unsupported shape: {default_shape}'
        self.default_shape = default_shape
        self.units = units
        self.disabled_counters = disabled_counters
        self.as_strings = as_strings
        self.measure_inference = measure_inference

    @abstractmethod
    def estimate(
        self, model: torch.nn.Module, resource_args: Dict[str, Any] = dict()
    ) -> Dict[str, float]:
        """Estimate the resources(flops/params/latency) of the given model.

        Args:
            model: The measured model.
            resource_args (Dict[str, float]): resources information.
            NOTE: resource_args have the same items() as the init cfgs.

        Returns:
            Dict[str, float]): A dict that containing resource results(flops,
                params and latency).
        """
        pass



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from mmrazor.registry import TASK_UTILS

try:
    from mmcls.models import ImageClassifier
except ImportError:
    from mmrazor.utils import get_placeholder
    ImageClassifier = get_placeholder('mmcls')


@TASK_UTILS.register_module()
class ImageClassifierPseudoLoss:
    """Calculate the pseudo loss to trace the topology of a `ImageClassifier`
    in MMClassification with `BackwardTracer`."""

    def __call__(self, model: ImageClassifier) -> torch.Tensor:
        pseudo_img = torch.rand(1, 3, 224, 224)
        pseudo_output = model(pseudo_img)
        return sum(pseudo_output)



# Copyright (c) OpenMMLab. All rights reserved.
from .autoslim_val_loop import AutoSlimValLoop
from .darts_loop import DartsEpochBasedTrainLoop, DartsIterBasedTrainLoop
from .distill_val_loop import SelfDistillValLoop, SingleTeacherDistillValLoop
from .evolution_search_loop import EvolutionSearchLoop
from .slimmable_val_loop import SlimmableValLoop
from .subnet_sampler_loop import GreedySamplerTrainLoop

__all__ = [
    'SingleTeacherDistillValLoop', 'DartsEpochBasedTrainLoop',
    'DartsIterBasedTrainLoop', 'SlimmableValLoop', 'EvolutionSearchLoop',
    'GreedySamplerTrainLoop', 'AutoSlimValLoop', 'SelfDistillValLoop'
]



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine.model import BaseModule


class BaseOP(BaseModule):
    """Base class for searchable operations.

    Args:
        in_channels (int): The input channels of the operation.
        out_channels (int): The output channels of the operation.
        stride (int): Stride of the operation. Defaults to 1.
    """

    def __init__(self, in_channels, out_channels, stride=1, **kwargs):
        super(BaseOP, self).__init__(**kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable, Dict

import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from mmrazor.models.mutables.base_mutable import BaseMutable
from mmrazor.registry import MODELS
from .dynamic_conv_mixins import (BigNasConvMixin, DynamicConvMixin,
                                  OFAConvMixin)


@MODELS.register_module()
class DynamicConv2d(nn.Conv2d, DynamicConvMixin):
    """Dynamic Conv2d OP.

    Note:
        Arguments for ``__init__`` of ``DynamicConv2d`` is totally same as
        :obj:`torch.nn.Conv2d`.

    Attributes:
        mutable_attrs (ModuleDict[str, BaseMutable]): Mutable attributes,
            such as `in_channels`. The key of the dict must in
            ``accepted_mutable_attrs``.
    """
    accepted_mutable_attrs = {'in_channels', 'out_channels'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # TODO
        # https://pytorch.org/docs/stable/_modules/torch/nn/modules/conv.html#Conv2d
        assert self.padding_mode == 'zeros'
        self.mutable_attrs: Dict[str, BaseMutable] = nn.ModuleDict()

    @classmethod
    def convert_from(cls, module: nn.Conv2d) -> 'DynamicConv2d':
        """Convert an instance of nn.Conv2d to a new instance of
        DynamicConv2d."""
        return cls(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=True if module.bias is not None else False,
            padding_mode=module.padding_mode)

    @property
    def conv_func(self) -> Callable:
        """The function that will be used in ``forward_mixin``."""
        return F.conv2d

    @property
    def static_op_factory(self):
        """Corresponding Pytorch OP."""
        return nn.Conv2d

    def forward(self, x: Tensor) -> Tensor:
        """Forward of dynamic conv2d OP."""
        return self.forward_mixin(x)


@MODELS.register_module()
class BigNasConv2d(nn.Conv2d, BigNasConvMixin):
    """Conv2d used in BigNas.

    Note:
        Arguments for ``__init__`` of ``DynamicConv2d`` is totally same as
        :obj:`torch.nn.Conv2d`.

    Attributes:
        mutable_attrs (ModuleDict[str, BaseMutable]): Mutable attributes,
            such as `in_channels`. The key of the dict must in
            ``accepted_mutable_attrs``.
    """
    accepted_mutable_attrs = {'in_channels', 'out_channels', 'kernel_size'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # TODO
        # https://pytorch.org/docs/stable/_modules/torch/nn/modules/conv.html#Conv2d
        assert self.padding_mode == 'zeros'
        self.mutable_attrs: Dict[str, BaseMutable] = nn.ModuleDict()

    @classmethod
    def convert_from(cls, module: nn.Conv2d) -> 'BigNasConv2d':
        """Convert an instance of `nn.Conv2d` to a new instance of
        `BigNasConv2d`."""
        return cls(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=True if module.bias is not None else False,
            padding_mode=module.padding_mode)

    @property
    def conv_func(self) -> Callable:
        """The function that will be used in ``forward_mixin``."""
        return F.conv2d

    @property
    def static_op_factory(self):
        """Corresponding Pytorch OP."""
        return nn.Conv2d

    def forward(self, x: Tensor) -> Tensor:
        """Forward of bignas' conv2d."""
        return self.forward_mixin(x)


@MODELS.register_module()
class OFAConv2d(nn.Conv2d, OFAConvMixin):
    """Conv2d used in `Once-for-All`.

    Refers to `Once-for-All: Train One Network and Specialize it for Efficient
    Deployment <http://arxiv.org/abs/1908.09791>`_.
    """
    """Dynamic Conv2d OP.

    Note:
        Arguments for ``__init__`` of ``OFAConv2d`` is totally same as
        :obj:`torch.nn.Conv2d`.

    Attributes:
        mutable_attrs (ModuleDict[str, BaseMutable]): Mutable attributes,
            such as `in_channels`. The key of the dict must in
            ``accepted_mutable_attrs``.
    """
    accepted_mutable_attrs = {'in_channels', 'out_channels'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # TODO
        # https://pytorch.org/docs/stable/_modules/torch/nn/modules/conv.html#Conv2d
        assert self.padding_mode == 'zeros'
        self.mutable_attrs: Dict[str, BaseMutable] = nn.ModuleDict()

    @classmethod
    def convert_from(cls, module: nn.Conv2d) -> 'OFAConv2d':
        """Convert an instance of `nn.Conv2d` to a new instance of
        `OFAConv2d`."""
        return cls(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=True if module.bias is not None else False,
            padding_mode=module.padding_mode)

    @property
    def conv_func(self) -> Callable:
        """The function that will be used in ``forward_mixin``."""
        return F.conv2d

    @property
    def static_op_factory(self):
        """Corresponding Pytorch OP."""
        return nn.Conv2d

    def forward(self, x: Tensor) -> Tensor:
        """Forward of OFA's conv2d."""
        return self.forward_mixin(x)



# Copyright (c) OpenMMLab. All rights reserved.
"""This module defines BaseNode and BaseGraph, which are used to model Directed
Acyclic Graph(DAG)"""
from collections import OrderedDict
from copy import copy
from typing import Any, Callable, Generic, Iterator, List, TypeVar

# BaseNode && BaseGraph


class BaseNode:
    """A single node in a graph."""

    def __init__(self, name: str, val: Any) -> None:
        """
        Args:
            name (str): name of the node.
            val (any): content of the node.
        """
        self.val = val
        self.name = name
        self.prev_nodes: List = []
        self.next_nodes: List = []

    # node operation

    def add_prev_node(self, node: 'BaseNode'):
        """add previous node."""
        if node not in self.prev_nodes:
            self.prev_nodes.append(node)
        if self not in node.next_nodes:
            node.next_nodes.append(self)

    def add_next_node(self, node: 'BaseNode'):
        """add next node."""
        if node not in self.next_nodes:
            self.next_nodes.append(node)
        if self not in node.prev_nodes:
            node.prev_nodes.append(self)

    @classmethod
    def copy_from(cls, node: 'BaseNode'):
        """Copy a node, and generate a new node with current node type."""
        return cls(node.name, node.val)

    # compare operation

    def __hash__(self) -> int:
        """Hash the node."""
        return hash((self.val, self.name))

    def __eq__(self, other):
        """Compare two nodes."""
        return self.val is other.val and self.name == other.name

    # other

    def __repr__(self) -> str:
        return self.name


BASENODE = TypeVar('BASENODE', bound=BaseNode)


class BaseGraph(Generic[BASENODE]):
    """A Directed Acyclic Graph(DAG)"""

    def __init__(self) -> None:
        super().__init__()
        self.nodes: OrderedDict[str, BASENODE] = OrderedDict()

    # graph operations

    @classmethod
    def copy_from(cls,
                  graph: 'BaseGraph',
                  node_converter: Callable = BaseNode.copy_from):
        """Copy a graph, and generate a new graph of the current class.

        Args:
            graph (BaseGraph): the graph to be copied.
            node_converter (Callable): a function that converts node,
            when coping graph.
        """
        old2new = {}
        new_graph = cls()
        # copy nodes
        for old in graph:
            old2new[old] = new_graph.add_or_find_node(node_converter(old))

        # connect
        for old in graph:
            for pre in old.prev_nodes:
                new_graph.connect(old2new[pre], old2new[old])
        return new_graph

    # node operations

    def add_or_find_node(self, node: BASENODE):
        """Add a node to the graph.

        If the node has exsited in the graph, the function will return the node
        recorded in the graph.
        """
        find = self.find_node(node)
        if find is not None:
            return find
        else:
            self.add_node(node)
            return node

    def find_node(self, node: BaseNode):
        """Find a node and return."""
        if node.name in self.nodes and node.val == self.nodes[node.name].val:
            return self.nodes[node.name]
        else:
            return None

    def add_node(self, node: BASENODE):
        """Add a node."""
        if node.name not in self.nodes:
            self.nodes[node.name] = node
        else:
            raise BaseException(f'{node.name} already exists in graph')

    def connect(self, pre_node: BASENODE, next_node: BASENODE):
        """Add an edge from pre_node to next_node."""
        assert pre_node in self and next_node in self
        pre_node.add_next_node(next_node)
        next_node.add_prev_node(pre_node)

    def disconnect(self, pre_node: BASENODE, next_node: BASENODE):
        """Remove the edge form pre_node to next_node."""
        assert pre_node in self and next_node in self
        if next_node in pre_node.next_nodes:
            pre_node.next_nodes.remove(next_node)
        if pre_node in next_node.prev_nodes:
            next_node.prev_nodes.remove(pre_node)

    def delete_node(self, node: BASENODE):
        """Delete a node with its related edges."""
        node = self.find_node(node)
        assert node is not None

        if len(node.prev_nodes) == 0:
            for next in copy(node.next_nodes):
                self.disconnect(node, next)
        elif len(node.next_nodes) == 0:
            for pre in copy(node.prev_nodes):
                self.disconnect(pre, node)
        elif len(node.prev_nodes) == 1:
            pre_node = node.prev_nodes[0]
            self.disconnect(pre_node, node)
            for next in copy(node.next_nodes):
                self.disconnect(node, next)
                self.connect(pre_node, next)
        elif len(node.next_nodes) == 1:
            next_node = node.next_nodes[0]
            self.disconnect(node, next_node)
            for pre in copy(node.prev_nodes):
                self.disconnect(pre, node)
                self.connect(pre, next_node)
        else:
            raise Exception(f'not delete {node}, \
                as it has more than one inputs and outputs')
        self.nodes.pop(node.name)

    # work as a collection

    def __iter__(self) -> Iterator[BASENODE]:
        """Traverse all nodes in the graph."""
        for x in self.nodes.values():
            yield x

    def __contains__(self, node: BASENODE) -> bool:
        """Check if a node is contained in the graph."""
        return node.name in self.nodes

    def __len__(self) -> int:
        """Number of nodes in the graph."""
        return len(self.nodes)

    # other

    def __repr__(self):
        res = f'Graph with {len(self)} nodes:\n'
        for node in self:
            res += '{0:<40} -> {1:^40} -> {2:<40}\n'.format(
                str(node.prev_nodes), node.__repr__(), str(node.next_nodes))
        return res

    # traverse

    def topo_traverse(self) -> Iterator[BASENODE]:
        """Traverse the graph in topologitcal order."""

        def _in_degree(graph: BaseGraph):
            degree = {}
            for name, node in graph.nodes.items():
                degree[name] = len(node.prev_nodes)
            return degree

        def find_zero_degree_node(in_degree):
            for node_name in in_degree:
                if in_degree[node_name] == 0:
                    return node_name
            return None

        in_degree = _in_degree(self)

        while len(in_degree) > 0:
            node_name = find_zero_degree_node(in_degree)  # visit the node
            in_degree.pop(node_name)
            yield self.nodes[node_name]
            for next in self.nodes[node_name].next_nodes:
                in_degree[next.name] -= 1

    def topo_sort(self):
        """Sort all node in topological order."""
        sorted_nodes = OrderedDict()
        for node in self.topo_traverse():
            sorted_nodes[node.name] = node
        self.nodes = sorted_nodes



_base_ = 'autoslim_mbv2_1.5x_slimmable_subnet_8xb256_in1k.py'

_channel_cfg_paths = 'https://download.openmmlab.com/mmrazor/v1/autoslim/autoslim_mbv2_subnet_8xb256_in1k_flops-320M_acc-72.73_20220715-9aa8f8ae_subnet_cfg.yaml'  # noqa: E501
model = dict(channel_cfg_paths=_channel_cfg_paths)



_base_ = ['./spos_shufflenet_supernet_8xb128_in1k.py']

model = dict(norm_training=True)

train_cfg = dict(
    _delete_=True,
    type='mmrazor.EvolutionSearchLoop',
    dataloader=_base_.val_dataloader,
    evaluator=_base_.val_evaluator,
    max_epochs=20,
    num_candidates=50,
    top_k=10,
    num_mutation=25,
    num_crossover=25,
    mutate_prob=0.1,
    flops_range=(0., 330 * 1e6),
    score_key='accuracy/top1')
